# Generated from \CSharpPreprocessorParser.g4 by ANTLR 4.10.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4, 1, 198, 123, 2, 0, 7, 0, 2, 1, 7, 1, 2, 2, 7, 2, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
        1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0,
        1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 3, 0, 38, 8, 0, 1, 0, 1, 0, 3, 0, 42, 8, 0, 1,
        0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 3, 0, 59,
        8, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 3, 0, 66, 8, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1,
        0, 3, 0, 76, 8, 0, 1, 1, 1, 1, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2,
        1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 3, 2, 96, 8, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1,
        2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 5, 2, 118, 8, 2, 10, 2, 12,
        2, 121, 9, 2, 1, 2, 0, 1, 4, 3, 0, 2, 4, 0, 1, 1, 1, 196, 196, 143, 0, 75, 1, 0, 0, 0, 2, 77,
        1, 0, 0, 0, 4, 95, 1, 0, 0, 0, 6, 7, 5, 183, 0, 0, 7, 8, 5, 195, 0, 0, 8, 9, 3, 2, 1, 0, 9, 10,
        6, 0, -1, 0, 10, 76, 1, 0, 0, 0, 11, 12, 5, 184, 0, 0, 12, 13, 5, 195, 0, 0, 13, 14, 3, 2,
        1, 0, 14, 15, 6, 0, -1, 0, 15, 76, 1, 0, 0, 0, 16, 17, 5, 52, 0, 0, 17, 18, 3, 4, 2, 0, 18,
        19, 3, 2, 1, 0, 19, 20, 6, 0, -
        1, 0, 20, 76, 1, 0, 0, 0, 21, 22, 5, 185, 0, 0, 22, 23, 3,
        4, 2, 0, 23, 24, 3, 2, 1, 0, 24, 25, 6, 0, -
        1, 0, 25, 76, 1, 0, 0, 0, 26, 27, 5, 36, 0, 0,
        27, 28, 3, 2, 1, 0, 28, 29, 6, 0, -
        1, 0, 29, 76, 1, 0, 0, 0, 30, 31, 5, 186, 0, 0, 31, 32,
        3, 2, 1, 0, 32, 33, 6, 0, -
        1, 0, 33, 76, 1, 0, 0, 0, 34, 41, 5, 187, 0, 0, 35, 37, 5, 182,
        0, 0, 36, 38, 5, 91, 0, 0, 37, 36, 1, 0, 0, 0, 37, 38, 1, 0, 0, 0, 38, 42, 1, 0, 0, 0, 39,
        42, 5, 30, 0, 0, 40, 42, 5, 194, 0, 0, 41, 35, 1, 0, 0, 0, 41, 39, 1, 0, 0, 0, 41, 40, 1,
        0, 0, 0, 42, 43, 1, 0, 0, 0, 43, 44, 3, 2, 1, 0, 44, 45, 6, 0, -
        1, 0, 45, 76, 1, 0, 0, 0, 46,
        47, 5, 188, 0, 0, 47, 48, 5, 197, 0, 0, 48, 49, 3, 2, 1, 0, 49, 50, 6, 0, -1, 0, 50, 76,
        1, 0, 0, 0, 51, 52, 5, 189, 0, 0, 52, 53, 5, 197, 0, 0, 53, 54, 3, 2, 1, 0, 54, 55, 6, 0,
        -1, 0, 55, 76, 1, 0, 0, 0, 56, 58, 5, 190, 0, 0, 57, 59, 5, 197, 0, 0, 58, 57, 1, 0, 0, 0,
        58, 59, 1, 0, 0, 0, 59, 60, 1, 0, 0, 0, 60, 61, 3, 2, 1, 0, 61, 62, 6, 0, -1, 0, 62, 76, 1,
        0, 0, 0, 63, 65, 5, 191, 0, 0, 64, 66, 5, 197, 0, 0, 65, 64, 1, 0, 0, 0, 65, 66, 1, 0, 0,
        0, 66, 67, 1, 0, 0, 0, 67, 68, 3, 2, 1, 0, 68, 69, 6, 0, -
        1, 0, 69, 76, 1, 0, 0, 0, 70, 71,
        5, 192, 0, 0, 71, 72, 5, 197, 0, 0, 72, 73, 3, 2, 1, 0, 73, 74, 6, 0, -1, 0, 74, 76, 1, 0,
        0, 0, 75, 6, 1, 0, 0, 0, 75, 11, 1, 0, 0, 0, 75, 16, 1, 0, 0, 0, 75, 21, 1, 0, 0, 0, 75, 26,
        1, 0, 0, 0, 75, 30, 1, 0, 0, 0, 75, 34, 1, 0, 0, 0, 75, 46, 1, 0, 0, 0, 75, 51, 1, 0, 0, 0,
        75, 56, 1, 0, 0, 0, 75, 63, 1, 0, 0, 0, 75, 70, 1, 0, 0, 0, 76, 1, 1, 0, 0, 0, 77, 78, 7, 0,
        0, 0, 78, 3, 1, 0, 0, 0, 79, 80, 6, 2, -
        1, 0, 80, 81, 5, 96, 0, 0, 81, 96, 6, 2, -1, 0, 82,
        83, 5, 42, 0, 0, 83, 96, 6, 2, -
        1, 0, 84, 85, 5, 195, 0, 0, 85, 96, 6, 2, -1, 0, 86, 87,
        5, 129, 0, 0, 87, 88, 3, 4, 2, 0, 88, 89, 5, 130, 0, 0, 89, 90, 6, 2, -1, 0, 90, 96, 1, 0,
        0, 0, 91, 92, 5, 143, 0, 0, 92, 93, 3, 4, 2, 5, 93, 94, 6, 2, -
        1, 0, 94, 96, 1, 0, 0, 0, 95,
        79, 1, 0, 0, 0, 95, 82, 1, 0, 0, 0, 95, 84, 1, 0, 0, 0, 95, 86, 1, 0, 0, 0, 95, 91, 1, 0, 0,
        0, 96, 119, 1, 0, 0, 0, 97, 98, 10, 4, 0, 0, 98, 99, 5, 156, 0, 0, 99, 100, 3, 4, 2, 5, 100,
        101, 6, 2, -1, 0, 101, 118, 1, 0, 0, 0, 102, 103, 10, 3, 0, 0, 103, 104, 5, 157, 0, 0,
        104, 105, 3, 4, 2, 4, 105, 106, 6, 2, -
        1, 0, 106, 118, 1, 0, 0, 0, 107, 108, 10, 2, 0,
        0, 108, 109, 5, 153, 0, 0, 109, 110, 3, 4, 2, 3, 110, 111, 6, 2, -1, 0, 111, 118, 1, 0,
        0, 0, 112, 113, 10, 1, 0, 0, 113, 114, 5, 154, 0, 0, 114, 115, 3, 4, 2, 2, 115, 116, 6,
        2, -1, 0, 116, 118, 1, 0, 0, 0, 117, 97, 1, 0, 0, 0, 117, 102, 1, 0, 0, 0, 117, 107, 1,
        0, 0, 0, 117, 112, 1, 0, 0, 0, 118, 121, 1, 0, 0, 0, 119, 117, 1, 0, 0, 0, 119, 120, 1,
        0, 0, 0, 120, 5, 1, 0, 0, 0, 121, 119, 1, 0, 0, 0, 8, 37, 41, 58, 65, 75, 95, 117, 119
    ]


class CSharpPreprocessorParser (Parser):

    grammarFileName = "CSharpPreprocessorParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = ["<INVALID>", "'\\u00EF\\u00BB\\u00BF'", "<INVALID>",
                    "'/***/'", "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "'#'", "'abstract'", "'add'", "'alias'", "'__arglist'",
                    "'as'", "'ascending'", "'async'", "'await'", "'base'",
                    "'bool'", "'break'", "'by'", "'byte'", "'case'", "'catch'",
                    "'char'", "'checked'", "'class'", "'const'", "'continue'",
                    "'decimal'", "'default'", "'delegate'", "'descending'",
                    "'do'", "'double'", "'dynamic'", "'else'", "'enum'",
                    "'equals'", "'event'", "'explicit'", "'extern'", "'False'",
                    "'finally'", "'fixed'", "'float'", "'for'", "'foreach'",
                    "'from'", "'get'", "'goto'", "'group'", "'if'", "'implicit'",
                    "'in'", "'int'", "'interface'", "'internal'", "'into'",
                    "'is'", "'join'", "'let'", "'lock'", "'long'", "'nameof'",
                    "'namespace'", "'new'", "'null'", "'object'", "'on'",
                    "'operator'", "'orderby'", "'out'", "'override'", "'params'",
                    "'partial'", "'private'", "'protected'", "'public'",
                    "'readonly'", "'ref'", "'remove'", "'return'", "'sbyte'",
                    "'sealed'", "'select'", "'set'", "'short'", "'sizeof'",
                    "'stackalloc'", "'static'", "'string'", "'struct'",
                    "'switch'", "'this'", "'throw'", "'true'", "'try'",
                    "'typeof'", "'uint'", "'ulong'", "'unchecked'", "'unmanaged'",
                    "'unsafe'", "'ushort'", "'using'", "'var'", "'virtual'",
                    "'void'", "'volatile'", "'when'", "'where'", "'while'",
                    "'yield'", "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "<INVALID>", "<INVALID>", "<INVALID>", "'{'", "'}'",
                    "'['", "']'", "'('", "')'", "'.'", "','", "':'", "';'",
                    "'+'", "'-'", "'*'", "'/'", "'%'", "'&'", "'|'", "'^'",
                    "'!'", "'~'", "'='", "'<'", "'>'", "'?'", "'::'", "'??'",
                    "'++'", "'--'", "'&&'", "'||'", "'->'", "'=='", "'!='",
                    "'<='", "'>='", "'+='", "'-='", "'*='", "'/='", "'%='",
                    "'&='", "'|='", "'^='", "'<<'", "'<<='", "'??='", "'..'",
                    "'{{'", "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>",
                    "<INVALID>", "<INVALID>", "'define'", "'undef'", "'elif'",
                    "'endif'", "'line'", "<INVALID>", "<INVALID>", "<INVALID>",
                    "<INVALID>", "<INVALID>", "<INVALID>", "'hidden'",
                    "<INVALID>", "<INVALID>", "<INVALID>", "'}}'"]

    symbolicNames = ["<INVALID>", "BYTE_ORDER_MARK", "SINGLE_LINE_DOC_COMMENT",
                     "EMPTY_DELIMITED_DOC_COMMENT", "DELIMITED_DOC_COMMENT",
                     "SINGLE_LINE_COMMENT", "DELIMITED_COMMENT", "WHITESPACES",
                     "SHARP", "ABSTRACT", "ADD", "ALIAS", "ARGLIST", "AS",
                     "ASCENDING", "ASYNC", "AWAIT", "BASE", "BOOL", "BREAK",
                     "BY", "BYTE", "CASE", "CATCH", "CHAR", "CHECKED",
                     "CLASS", "CONST", "CONTINUE", "DECIMAL", "DEFAULT",
                     "DELEGATE", "DESCENDING", "DO", "DOUBLE", "DYNAMIC",
                     "ELSE", "ENUM", "EQUALS", "EVENT", "EXPLICIT", "EXTERN",
                     "False", "FINALLY", "FIXED", "FLOAT", "FOR", "FOREACH",
                     "FROM", "GET", "GOTO", "GROUP", "IF", "IMPLICIT",
                     "IN", "INT", "INTERFACE", "INTERNAL", "INTO", "IS",
                     "JOIN", "LET", "LOCK", "LONG", "NAMEOF", "NAMESPACE",
                     "NEW", "NULL_", "OBJECT", "ON", "OPERATOR", "ORDERBY",
                     "OUT", "OVERRIDE", "PARAMS", "PARTIAL", "PRIVATE",
                     "PROTECTED", "PUBLIC", "READONLY", "REF", "REMOVE",
                     "RETURN", "SBYTE", "SEALED", "SELECT", "SET", "SHORT",
                     "SIZEOF", "STACKALLOC", "STATIC", "STRING", "STRUCT",
                     "SWITCH", "THIS", "THROW", "TRUE", "TRY", "TYPEOF",
                     "UINT", "ULONG", "UNCHECKED", "UNMANAGED", "UNSAFE",
                     "USHORT", "USING", "VAR", "VIRTUAL", "VOID", "VOLATILE",
                     "WHEN", "WHERE", "WHILE", "YIELD", "IDENTIFIER", "LITERAL_ACCESS",
                     "INTEGER_LITERAL", "HEX_INTEGER_LITERAL", "BIN_INTEGER_LITERAL",
                     "REAL_LITERAL", "CHARACTER_LITERAL", "REGULAR_STRING",
                     "VERBATIUM_STRING", "INTERPOLATED_REGULAR_STRING_START",
                     "INTERPOLATED_VERBATIUM_STRING_START", "OPEN_BRACE",
                     "CLOSE_BRACE", "OPEN_BRACKET", "CLOSE_BRACKET", "OPEN_PARENS",
                     "CLOSE_PARENS", "DOT", "COMMA", "COLON", "SEMICOLON",
                     "PLUS", "MINUS", "STAR", "DIV", "PERCENT", "AMP",
                     "BITWISE_OR", "CARET", "BANG", "TILDE", "ASSIGNMENT",
                     "LT", "GT", "INTERR", "DOUBLE_COLON", "OP_COALESCING",
                     "OP_INC", "OP_DEC", "OP_AND", "OP_OR", "OP_PTR", "OP_EQ",
                     "OP_NE", "OP_LE", "OP_GE", "OP_ADD_ASSIGNMENT", "OP_SUB_ASSIGNMENT",
                     "OP_MULT_ASSIGNMENT", "OP_DIV_ASSIGNMENT", "OP_MOD_ASSIGNMENT",
                     "OP_AND_ASSIGNMENT", "OP_OR_ASSIGNMENT", "OP_XOR_ASSIGNMENT",
                     "OP_LEFT_SHIFT", "OP_LEFT_SHIFT_ASSIGNMENT", "OP_COALESCING_ASSIGNMENT",
                     "OP_RANGE", "DOUBLE_CURLY_INSIDE", "OPEN_BRACE_INSIDE",
                     "REGULAR_CHAR_INSIDE", "VERBATIUM_DOUBLE_QUOTE_INSIDE",
                     "DOUBLE_QUOTE_INSIDE", "REGULAR_STRING_INSIDE", "VERBATIUM_INSIDE_STRING",
                     "CLOSE_BRACE_INSIDE", "FORMAT_STRING", "DIRECTIVE_WHITESPACES",
                     "DIGITS", "DEFINE", "UNDEF", "ELIF", "ENDIF", "LINE",
                     "ERROR", "WARNING", "REGION", "ENDREGION", "PRAGMA",
                     "NULLABLE", "DIRECTIVE_HIDDEN", "CONDITIONAL_SYMBOL",
                     "DIRECTIVE_NEW_LINE", "TEXT", "DOUBLE_CURLY_CLOSE_INSIDE"]

    RULE_preprocessor_directive = 0
    RULE_directive_new_line_or_sharp = 1
    RULE_preprocessor_expression = 2

    ruleNames = ["preprocessor_directive", "directive_new_line_or_sharp",
                 "preprocessor_expression"]

    EOF = Token.EOF
    BYTE_ORDER_MARK = 1
    SINGLE_LINE_DOC_COMMENT = 2
    EMPTY_DELIMITED_DOC_COMMENT = 3
    DELIMITED_DOC_COMMENT = 4
    SINGLE_LINE_COMMENT = 5
    DELIMITED_COMMENT = 6
    WHITESPACES = 7
    SHARP = 8
    ABSTRACT = 9
    ADD = 10
    ALIAS = 11
    ARGLIST = 12
    AS = 13
    ASCENDING = 14
    ASYNC = 15
    AWAIT = 16
    BASE = 17
    BOOL = 18
    BREAK = 19
    BY = 20
    BYTE = 21
    CASE = 22
    CATCH = 23
    CHAR = 24
    CHECKED = 25
    CLASS = 26
    CONST = 27
    CONTINUE = 28
    DECIMAL = 29
    DEFAULT = 30
    DELEGATE = 31
    DESCENDING = 32
    DO = 33
    DOUBLE = 34
    DYNAMIC = 35
    ELSE = 36
    ENUM = 37
    EQUALS = 38
    EVENT = 39
    EXPLICIT = 40
    EXTERN = 41
    FALSE = 42
    FINALLY = 43
    FIXED = 44
    FLOAT = 45
    FOR = 46
    FOREACH = 47
    FROM = 48
    GET = 49
    GOTO = 50
    GROUP = 51
    IF = 52
    IMPLICIT = 53
    IN = 54
    INT = 55
    INTERFACE = 56
    INTERNAL = 57
    INTO = 58
    IS = 59
    JOIN = 60
    LET = 61
    LOCK = 62
    LONG = 63
    NAMEOF = 64
    NAMESPACE = 65
    NEW = 66
    NULL_ = 67
    OBJECT = 68
    ON = 69
    OPERATOR = 70
    ORDERBY = 71
    OUT = 72
    OVERRIDE = 73
    PARAMS = 74
    PARTIAL = 75
    PRIVATE = 76
    PROTECTED = 77
    PUBLIC = 78
    READONLY = 79
    REF = 80
    REMOVE = 81
    RETURN = 82
    SBYTE = 83
    SEALED = 84
    SELECT = 85
    SET = 86
    SHORT = 87
    SIZEOF = 88
    STACKALLOC = 89
    STATIC = 90
    STRING = 91
    STRUCT = 92
    SWITCH = 93
    THIS = 94
    THROW = 95
    TRUE = 96
    TRY = 97
    TYPEOF = 98
    UINT = 99
    ULONG = 100
    UNCHECKED = 101
    UNMANAGED = 102
    UNSAFE = 103
    USHORT = 104
    USING = 105
    VAR = 106
    VIRTUAL = 107
    VOID = 108
    VOLATILE = 109
    WHEN = 110
    WHERE = 111
    WHILE = 112
    YIELD = 113
    IDENTIFIER = 114
    LITERAL_ACCESS = 115
    INTEGER_LITERAL = 116
    HEX_INTEGER_LITERAL = 117
    BIN_INTEGER_LITERAL = 118
    REAL_LITERAL = 119
    CHARACTER_LITERAL = 120
    REGULAR_STRING = 121
    VERBATIUM_STRING = 122
    INTERPOLATED_REGULAR_STRING_START = 123
    INTERPOLATED_VERBATIUM_STRING_START = 124
    OPEN_BRACE = 125
    CLOSE_BRACE = 126
    OPEN_BRACKET = 127
    CLOSE_BRACKET = 128
    OPEN_PARENS = 129
    CLOSE_PARENS = 130
    DOT = 131
    COMMA = 132
    COLON = 133
    SEMICOLON = 134
    PLUS = 135
    MINUS = 136
    STAR = 137
    DIV = 138
    PERCENT = 139
    AMP = 140
    BITWISE_OR = 141
    CARET = 142
    BANG = 143
    TILDE = 144
    ASSIGNMENT = 145
    LT = 146
    GT = 147
    INTERR = 148
    DOUBLE_COLON = 149
    OP_COALESCING = 150
    OP_INC = 151
    OP_DEC = 152
    OP_AND = 153
    OP_OR = 154
    OP_PTR = 155
    OP_EQ = 156
    OP_NE = 157
    OP_LE = 158
    OP_GE = 159
    OP_ADD_ASSIGNMENT = 160
    OP_SUB_ASSIGNMENT = 161
    OP_MULT_ASSIGNMENT = 162
    OP_DIV_ASSIGNMENT = 163
    OP_MOD_ASSIGNMENT = 164
    OP_AND_ASSIGNMENT = 165
    OP_OR_ASSIGNMENT = 166
    OP_XOR_ASSIGNMENT = 167
    OP_LEFT_SHIFT = 168
    OP_LEFT_SHIFT_ASSIGNMENT = 169
    OP_COALESCING_ASSIGNMENT = 170
    OP_RANGE = 171
    DOUBLE_CURLY_INSIDE = 172
    OPEN_BRACE_INSIDE = 173
    REGULAR_CHAR_INSIDE = 174
    VERBATIUM_DOUBLE_QUOTE_INSIDE = 175
    DOUBLE_QUOTE_INSIDE = 176
    REGULAR_STRING_INSIDE = 177
    VERBATIUM_INSIDE_STRING = 178
    CLOSE_BRACE_INSIDE = 179
    FORMAT_STRING = 180
    DIRECTIVE_WHITESPACES = 181
    DIGITS = 182
    DEFINE = 183
    UNDEF = 184
    ELIF = 185
    ENDIF = 186
    LINE = 187
    ERROR = 188
    WARNING = 189
    REGION = 190
    ENDREGION = 191
    PRAGMA = 192
    NULLABLE = 193
    DIRECTIVE_HIDDEN = 194
    CONDITIONAL_SYMBOL = 195
    DIRECTIVE_NEW_LINE = 196
    TEXT = 197
    DOUBLE_CURLY_CLOSE_INSIDE = 198

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None

    conditions = [True]
    ConditionalSymbols = {"DEBUG"}

    def allConditions(self):
        for condition in CSharpPreprocessorParser.conditions:
            if not condition:
                return False
        return True

    class Preprocessor_directiveContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.value = None

        def getRuleIndex(self):
            return CSharpPreprocessorParser.RULE_preprocessor_directive

        def copyFrom(self, ctx: ParserRuleContext):
            super().copyFrom(ctx)
            self.value = ctx.value

    class PreprocessorDiagnosticContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self.copyFrom(ctx)

        def ERROR(self):
            return self.getToken(CSharpPreprocessorParser.ERROR, 0)

        def TEXT(self):
            return self.getToken(CSharpPreprocessorParser.TEXT, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def WARNING(self):
            return self.getToken(CSharpPreprocessorParser.WARNING, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorDiagnostic"):
                listener.enterPreprocessorDiagnostic(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorDiagnostic"):
                listener.exitPreprocessorDiagnostic(self)

    class PreprocessorRegionContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self.copyFrom(ctx)

        def REGION(self):
            return self.getToken(CSharpPreprocessorParser.REGION, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def TEXT(self):
            return self.getToken(CSharpPreprocessorParser.TEXT, 0)

        def ENDREGION(self):
            return self.getToken(CSharpPreprocessorParser.ENDREGION, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorRegion"):
                listener.enterPreprocessorRegion(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorRegion"):
                listener.exitPreprocessorRegion(self)

    class PreprocessorDeclarationContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self._CONDITIONAL_SYMBOL = None  # Token
            self.copyFrom(ctx)

        def DEFINE(self):
            return self.getToken(CSharpPreprocessorParser.DEFINE, 0)

        def CONDITIONAL_SYMBOL(self):
            return self.getToken(CSharpPreprocessorParser.CONDITIONAL_SYMBOL, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def UNDEF(self):
            return self.getToken(CSharpPreprocessorParser.UNDEF, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorDeclaration"):
                listener.enterPreprocessorDeclaration(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorDeclaration"):
                listener.exitPreprocessorDeclaration(self)

    class PreprocessorConditionalContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self.expr = None  # Preprocessor_expressionContext
            self.copyFrom(ctx)

        def IF(self):
            return self.getToken(CSharpPreprocessorParser.IF, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def preprocessor_expression(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Preprocessor_expressionContext, 0)

        def ELIF(self):
            return self.getToken(CSharpPreprocessorParser.ELIF, 0)

        def ELSE(self):
            return self.getToken(CSharpPreprocessorParser.ELSE, 0)

        def ENDIF(self):
            return self.getToken(CSharpPreprocessorParser.ENDIF, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorConditional"):
                listener.enterPreprocessorConditional(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorConditional"):
                listener.exitPreprocessorConditional(self)

    class PreprocessorPragmaContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self.copyFrom(ctx)

        def PRAGMA(self):
            return self.getToken(CSharpPreprocessorParser.PRAGMA, 0)

        def TEXT(self):
            return self.getToken(CSharpPreprocessorParser.TEXT, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorPragma"):
                listener.enterPreprocessorPragma(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorPragma"):
                listener.exitPreprocessorPragma(self)

    class PreprocessorLineContext(Preprocessor_directiveContext):

        # actually a CSharpPreprocessorParser.Preprocessor_directiveContext
        def __init__(self, parser, ctx: ParserRuleContext):
            super().__init__(parser)
            self.copyFrom(ctx)

        def LINE(self):
            return self.getToken(CSharpPreprocessorParser.LINE, 0)

        def directive_new_line_or_sharp(self):
            return self.getTypedRuleContext(CSharpPreprocessorParser.Directive_new_line_or_sharpContext, 0)

        def DIGITS(self):
            return self.getToken(CSharpPreprocessorParser.DIGITS, 0)

        def DEFAULT(self):
            return self.getToken(CSharpPreprocessorParser.DEFAULT, 0)

        def DIRECTIVE_HIDDEN(self):
            return self.getToken(CSharpPreprocessorParser.DIRECTIVE_HIDDEN, 0)

        def STRING(self):
            return self.getToken(CSharpPreprocessorParser.STRING, 0)

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessorLine"):
                listener.enterPreprocessorLine(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessorLine"):
                listener.exitPreprocessorLine(self)

    def preprocessor_directive(self):

        localctx = CSharpPreprocessorParser.Preprocessor_directiveContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_preprocessor_directive)
        self._la = 0  # Token type
        try:
            self.state = 75
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [CSharpPreprocessorParser.DEFINE]:
                localctx = CSharpPreprocessorParser.PreprocessorDeclarationContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 6
                self.match(CSharpPreprocessorParser.DEFINE)
                self.state = 7
                localctx._CONDITIONAL_SYMBOL = self.match(
                    CSharpPreprocessorParser.CONDITIONAL_SYMBOL)
                self.state = 8
                self.directive_new_line_or_sharp()
                ConditionalSymbols.add(
                    (None if localctx._CONDITIONAL_SYMBOL is None else localctx._CONDITIONAL_SYMBOL.text))
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.UNDEF]:
                localctx = CSharpPreprocessorParser.PreprocessorDeclarationContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 11
                self.match(CSharpPreprocessorParser.UNDEF)
                self.state = 12
                localctx._CONDITIONAL_SYMBOL = self.match(
                    CSharpPreprocessorParser.CONDITIONAL_SYMBOL)
                self.state = 13
                self.directive_new_line_or_sharp()
                ConditionalSymbols.remove(
                    (None if localctx._CONDITIONAL_SYMBOL is None else localctx._CONDITIONAL_SYMBOL.text))
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.IF]:
                localctx = CSharpPreprocessorParser.PreprocessorConditionalContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 16
                self.match(CSharpPreprocessorParser.IF)
                self.state = 17
                localctx.expr = self.preprocessor_expression(0)
                self.state = 18
                self.directive_new_line_or_sharp()
                # conditions.push(localctx.expr.value.equals("true"))
                localctx.value = localctx.expr.value.equals(
                    "true") and self.allConditions()
                pass
            elif token in [CSharpPreprocessorParser.ELIF]:
                localctx = CSharpPreprocessorParser.PreprocessorConditionalContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 21
                self.match(CSharpPreprocessorParser.ELIF)
                self.state = 22
                localctx.expr = self.preprocessor_expression(0)
                self.state = 23
                self.directive_new_line_or_sharp()
                if not conditions.peek():
                    conditions.pop()
                    localctx.value = localctx.expr.value.equals(
                        "true") and allConditions()
                    # conditions.push(localctx.expr.value.equals("true")) else localctx.value = False
                pass
            elif token in [CSharpPreprocessorParser.ELSE]:
                localctx = CSharpPreprocessorParser.PreprocessorConditionalContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 26
                self.match(CSharpPreprocessorParser.ELSE)
                self.state = 27
                self.directive_new_line_or_sharp()
                if not conditions.peek():
                    conditions.pop()
                    localctx.value = true and allConditions()
                    # conditions.push(true)
                else:
                    localctx.value = False
                pass
            elif token in [CSharpPreprocessorParser.ENDIF]:
                localctx = CSharpPreprocessorParser.PreprocessorConditionalContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 30
                self.match(CSharpPreprocessorParser.ENDIF)
                self.state = 31
                self.directive_new_line_or_sharp()
                conditions.pop()
                localctx.value = conditions.peek()
                pass
            elif token in [CSharpPreprocessorParser.LINE]:
                localctx = CSharpPreprocessorParser.PreprocessorLineContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 7)
                self.state = 34
                self.match(CSharpPreprocessorParser.LINE)
                self.state = 41
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [CSharpPreprocessorParser.DIGITS]:
                    self.state = 35
                    self.match(CSharpPreprocessorParser.DIGITS)
                    self.state = 37
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la == CSharpPreprocessorParser.STRING:
                        self.state = 36
                        self.match(CSharpPreprocessorParser.STRING)

                    pass
                elif token in [CSharpPreprocessorParser.DEFAULT]:
                    self.state = 39
                    self.match(CSharpPreprocessorParser.DEFAULT)
                    pass
                elif token in [CSharpPreprocessorParser.DIRECTIVE_HIDDEN]:
                    self.state = 40
                    self.match(CSharpPreprocessorParser.DIRECTIVE_HIDDEN)
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 43
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.ERROR]:
                localctx = CSharpPreprocessorParser.PreprocessorDiagnosticContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 8)
                self.state = 46
                self.match(CSharpPreprocessorParser.ERROR)
                self.state = 47
                self.match(CSharpPreprocessorParser.TEXT)
                self.state = 48
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.WARNING]:
                localctx = CSharpPreprocessorParser.PreprocessorDiagnosticContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 9)
                self.state = 51
                self.match(CSharpPreprocessorParser.WARNING)
                self.state = 52
                self.match(CSharpPreprocessorParser.TEXT)
                self.state = 53
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.REGION]:
                localctx = CSharpPreprocessorParser.PreprocessorRegionContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 10)
                self.state = 56
                self.match(CSharpPreprocessorParser.REGION)
                self.state = 58
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == CSharpPreprocessorParser.TEXT:
                    self.state = 57
                    self.match(CSharpPreprocessorParser.TEXT)

                self.state = 60
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.ENDREGION]:
                localctx = CSharpPreprocessorParser.PreprocessorRegionContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 11)
                self.state = 63
                self.match(CSharpPreprocessorParser.ENDREGION)
                self.state = 65
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == CSharpPreprocessorParser.TEXT:
                    self.state = 64
                    self.match(CSharpPreprocessorParser.TEXT)

                self.state = 67
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            elif token in [CSharpPreprocessorParser.PRAGMA]:
                localctx = CSharpPreprocessorParser.PreprocessorPragmaContext(
                    self, localctx)
                self.enterOuterAlt(localctx, 12)
                self.state = 70
                self.match(CSharpPreprocessorParser.PRAGMA)
                self.state = 71
                self.match(CSharpPreprocessorParser.TEXT)
                self.state = 72
                self.directive_new_line_or_sharp()
                localctx.value = allConditions()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Directive_new_line_or_sharpContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def DIRECTIVE_NEW_LINE(self):
            return self.getToken(CSharpPreprocessorParser.DIRECTIVE_NEW_LINE, 0)

        def EOF(self):
            return self.getToken(CSharpPreprocessorParser.EOF, 0)

        def getRuleIndex(self):
            return CSharpPreprocessorParser.RULE_directive_new_line_or_sharp

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterDirective_new_line_or_sharp"):
                listener.enterDirective_new_line_or_sharp(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitDirective_new_line_or_sharp"):
                listener.exitDirective_new_line_or_sharp(self)

    def directive_new_line_or_sharp(self):

        localctx = CSharpPreprocessorParser.Directive_new_line_or_sharpContext(
            self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_directive_new_line_or_sharp)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 77
            _la = self._input.LA(1)
            if not (_la == CSharpPreprocessorParser.EOF or _la == CSharpPreprocessorParser.DIRECTIVE_NEW_LINE):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Preprocessor_expressionContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent: ParserRuleContext = None, invokingState: int = -1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.value = None
            self.expr1 = None  # Preprocessor_expressionContext
            self._CONDITIONAL_SYMBOL = None  # Token
            self.expr = None  # Preprocessor_expressionContext
            self.expr2 = None  # Preprocessor_expressionContext

        def TRUE(self):
            return self.getToken(CSharpPreprocessorParser.TRUE, 0)

        def FALSE(self):
            return self.getToken(CSharpPreprocessorParser.FALSE, 0)

        def CONDITIONAL_SYMBOL(self):
            return self.getToken(CSharpPreprocessorParser.CONDITIONAL_SYMBOL, 0)

        def OPEN_PARENS(self):
            return self.getToken(CSharpPreprocessorParser.OPEN_PARENS, 0)

        def CLOSE_PARENS(self):
            return self.getToken(CSharpPreprocessorParser.CLOSE_PARENS, 0)

        def preprocessor_expression(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(CSharpPreprocessorParser.Preprocessor_expressionContext)
            else:
                return self.getTypedRuleContext(CSharpPreprocessorParser.Preprocessor_expressionContext, i)

        def BANG(self):
            return self.getToken(CSharpPreprocessorParser.BANG, 0)

        def OP_EQ(self):
            return self.getToken(CSharpPreprocessorParser.OP_EQ, 0)

        def OP_NE(self):
            return self.getToken(CSharpPreprocessorParser.OP_NE, 0)

        def OP_AND(self):
            return self.getToken(CSharpPreprocessorParser.OP_AND, 0)

        def OP_OR(self):
            return self.getToken(CSharpPreprocessorParser.OP_OR, 0)

        def getRuleIndex(self):
            return CSharpPreprocessorParser.RULE_preprocessor_expression

        def enterRule(self, listener: ParseTreeListener):
            if hasattr(listener, "enterPreprocessor_expression"):
                listener.enterPreprocessor_expression(self)

        def exitRule(self, listener: ParseTreeListener):
            if hasattr(listener, "exitPreprocessor_expression"):
                listener.exitPreprocessor_expression(self)

    def preprocessor_expression(self, _p: int = 0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = CSharpPreprocessorParser.Preprocessor_expressionContext(
            self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(
            localctx, 4, self.RULE_preprocessor_expression, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [CSharpPreprocessorParser.TRUE]:
                self.state = 80
                self.match(CSharpPreprocessorParser.TRUE)
                localctx.value = "true"
                pass
            elif token in [CSharpPreprocessorParser.FALSE]:
                self.state = 82
                self.match(CSharpPreprocessorParser.FALSE)
                localctx.value = "False"
                pass
            elif token in [CSharpPreprocessorParser.CONDITIONAL_SYMBOL]:
                self.state = 84
                localctx._CONDITIONAL_SYMBOL = self.match(
                    CSharpPreprocessorParser.CONDITIONAL_SYMBOL)
                localctx.value = ConditionalSymbols.contains(
                    None if localctx._CONDITIONAL_SYMBOL is None else localctx._CONDITIONAL_SYMBOL.text)
                pass
            elif token in [CSharpPreprocessorParser.OPEN_PARENS]:
                self.state = 86
                self.match(CSharpPreprocessorParser.OPEN_PARENS)
                self.state = 87
                localctx.expr = self.preprocessor_expression(0)
                self.state = 88
                self.match(CSharpPreprocessorParser.CLOSE_PARENS)
                localctx.value = localctx.expr.value
                pass
            elif token in [CSharpPreprocessorParser.BANG]:
                self.state = 91
                self.match(CSharpPreprocessorParser.BANG)
                self.state = 92
                localctx.expr = self.preprocessor_expression(5)
                localctx.value = localctx.expr.value.equals("true")
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 119
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 117
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(
                        self._input, 6, self._ctx)
                    if la_ == 1:
                        localctx = CSharpPreprocessorParser.Preprocessor_expressionContext(
                            self, _parentctx, _parentState)
                        localctx.expr1 = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_preprocessor_expression)
                        self.state = 97
                        if not self.precpred(self._ctx, 4):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 4)")
                        self.state = 98
                        self.match(CSharpPreprocessorParser.OP_EQ)
                        self.state = 99
                        localctx.expr2 = self.preprocessor_expression(5)
                        localctx.value = localctx.expr1.value == localctx.expr2.value
                        pass

                    elif la_ == 2:
                        localctx = CSharpPreprocessorParser.Preprocessor_expressionContext(
                            self, _parentctx, _parentState)
                        localctx.expr1 = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_preprocessor_expression)
                        self.state = 102
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 3)")
                        self.state = 103
                        self.match(CSharpPreprocessorParser.OP_NE)
                        self.state = 104
                        localctx.expr2 = self.preprocessor_expression(4)
                        localctx.value = localctx.expr1.value != localctx.expr2.value
                        pass

                    elif la_ == 3:
                        localctx = CSharpPreprocessorParser.Preprocessor_expressionContext(
                            self, _parentctx, _parentState)
                        localctx.expr1 = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_preprocessor_expression)
                        self.state = 107
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 2)")
                        self.state = 108
                        self.match(CSharpPreprocessorParser.OP_AND)
                        self.state = 109
                        localctx.expr2 = self.preprocessor_expression(3)
                        localctx.value = localctx.expr1.value.equals(
                            "true") and localctx.expr2.value.equals("true")
                        pass

                    elif la_ == 4:
                        localctx = CSharpPreprocessorParser.Preprocessor_expressionContext(
                            self, _parentctx, _parentState)
                        localctx.expr1 = _prevctx
                        self.pushNewRecursionContext(
                            localctx, _startState, self.RULE_preprocessor_expression)
                        self.state = 112
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(
                                self, "self.precpred(self._ctx, 1)")
                        self.state = 113
                        self.match(CSharpPreprocessorParser.OP_OR)
                        self.state = 114
                        localctx.expr2 = self.preprocessor_expression(2)
                        localctx.value = localctx.expr1.value.equals(
                            "true") or localctx.expr2.value.equals("true")
                        pass

                self.state = 121
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    def sempred(self, localctx: RuleContext, ruleIndex: int, predIndex: int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[2] = self.preprocessor_expression_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def preprocessor_expression_sempred(self, localctx: Preprocessor_expressionContext, predIndex: int):
        if predIndex == 0:
            return self.precpred(self._ctx, 4)

        if predIndex == 1:
            return self.precpred(self._ctx, 3)

        if predIndex == 2:
            return self.precpred(self._ctx, 2)

        if predIndex == 3:
            return self.precpred(self._ctx, 1)



# -*- coding: utf-8 -*-
"""
@file
@brief Various functions about :epkg:`matplotlib`.
"""
import sys


def mpl_switch_style(style="ggplot"):
    """
    Changes the graph style.

    :param style: see `Customizing plots with style sheets
        <http://matplotlib.org/users/style_sheets.html>`_
    """
    # This import was moved here because ths code is executed when
    # the module is imported and for some reasons, it overides some of the settings
    # sphinx is doing and graphs are not part of the documentation but show up
    # in a separate window.
    if "plt" not in sys.modules:
        import matplotlib.pyplot as plt  # pylint: disable=C0415
    plt.style.use(style)



"""
@file

@brief @see cl Database
"""

import re
import copy


class DatabaseJoinGroup:

    """
    This class is not neant to be working alone.
    It contains functions for a database able to build
    SQL requests for frequent needs such as join SQL requests.
    @see cl Database
    """

    class JoinTreeNode:

        """define a node meant to be included in a graph to define a big join
        """

        def __init__(self, table,
                     parent_key=None,
                     key=None,
                     where=None,
                     prefix=None,
                     avoid_prefix=False):
            """constructor
            this node defines a join on two tables (parent_table, table)
            on two keys (parent_key, key). The keys can be tuple or string.

            @param      table           table name
            @param      parent_key      None if it is the root
            @param      key             None if it is the root
            @param      where           clause where
                                                where is a where clause defined as a dictionary: example:

                                                @code
                                                {   "field": ("==", value),
                                                    ("table","field"): (">=", value) }
                                                @endcode

                                                You may add field not connected to a table,
                                                they will not taken into account.
            @param      prefix          add a prefix, avoid different fields collide
            @param      avoid_prefix    avoid using a prefix to build SQL queries, use syntax ``( ... ) AS ...``
            """
            self.table = table
            self.parent_key = parent_key
            self.key = key
            self.where = {}
            self.predecessor = None
            self.successor = []
            self.prefix = None
            self._count_as = 0
            self._avoid_prefix = avoid_prefix
            if self.parent_key is None and self.key is not None:
                raise KeyError(  # pragma: no cover
                    "parent_key is missing")

            if where is not None:
                if not isinstance(where, dict):
                    raise TypeError(  # pragma: no cover
                        "parameter where: only dict or None expected (not %s)" % str(
                            type(where)))
                for k, v in where.items():
                    if not isinstance(k, tuple):
                        k = tuple(k.split("."))
                    if len(k) == 1:
                        self.where[k[0]] = v
                    elif len(k) == 2:
                        if k[-2] == self.table:
                            self.where[k[-1]] = v
                    else:
                        raise ValueError(  # pragma: no cover
                            "not able to deal with than clause %s:%s" %
                            (str(k), str(v)))

        def __str__(self):
            """usual
            """
            mes = ["*nb succ: %d" % len(self.successor)]
            for k, v in sorted(self.__dict__.items()):
                if k not in ["table", "parent_key", "where", "key"]:
                    continue
                s = k + " " * (12 - len(k))
                s += str(v)
                mes.append(s)
            i = 0
            for n in self.successor:
                r = str(n)
                li = r.split("\n")
                li = ["    " + s2 for s2 in li]
                r = "\n".join(li)
                mes.append("node %d" % i)
                mes.append(r)
                i += 1
            return "\n".join(mes) + "\n"

        def append(self, n):
            """add a successor
            @param      n       new successor
            """
            if n.predecessor is not None:
                raise ValueError(  # pragma: no cover
                    "This node was already added in another part of the tree. You must duplicate it.")
            self.successor.append(n)
            n.predecessor = self

        def get_nb_successor(self):
            """
            @return the number of successors
            """
            return len(self.successor)

        def check_prefix(self, nb=-1):
            """
            @param      nb      index of this node is the predecessor list of successor
            check the prefixes, all one if there is none
            """
            if self._avoid_prefix:
                self.PREFIX = ""
                return
            if "PREFIX" in self.__dict__:
                return
            if self.prefix is None:
                if nb == -1:
                    self.prefix = ""
                else:
                    self.prefix = chr(97 + nb)
            for i, n in enumerate(self.successor):
                n.check_prefix(i)
            self.PREFIX = "" if self.prefix is None else self.prefix
            self.PREFIX = self._build_predecessor_prefix() + self.PREFIX

        def _build_predecessor_prefix(self):
            """
            private method
            """
            if self._avoid_prefix:
                return ""
            r = ""
            n = self.predecessor
            while n is not None:
                r += n.prefix
                n = n.predecessor
            return r

        def clean(self):
            """
            remove all sql,fields members
            """
            if "SELECT" in self.__dict__:
                del self.__dict__["SELECT"]
                del self.__dict__["FIELDS"]
                del self.__dict__["PREFIX"]
            for n in self.successor:
                n.clean()

        def _in_select(self, db):
            """return the SQL select on the table
            @param      db  database
            @return         list of tuple (fieldas, table, field, which), where
            """
            fields = db.get_table_columns_list(self.table)
            where = {}
            for f, t in fields:
                if f in self.where:
                    where["%s.%s" % (self.table, f)] = self.where[f]
                elif (self.table, f) in self.where:
                    where["%s.%s" % (self.table, f)] = self.where
                elif "." in f:
                    table, z = f.split(".")[-2:]
                    if table == self.table and z in self.where:
                        where["%s.%s" % (table, z)] = self.where[z]

            prefix = self.PREFIX
            fas = []
            for f, t in fields:
                fas.append((prefix + f, self.table, f, True))
            return fas, where

        def _build_select(self, db, fas, where, tfrom=None):
            """build a select SQL request
            @param      db      database
            @param      fas     list of tuple table,f,fas
            @param      where   where clause
            @param      tfrom   from clause, if None, --> self.table
            @return             string
            """
            lines = []
            lines.append("SELECT")
            alkey = True
            if len(fas) > 0:
                mx = max([len(s[0]) for s in fas]) + 1
                name_changed = 0
                for well in fas:
                    fn, t, fo = well[:3]
                    if fn == self.key:
                        if alkey:
                            alkey = False
                            doit = True
                        else:
                            doit = False
                    else:
                        doit = True

                    if doit:
                        s = " " * (mx - len(fn))
                        if len(well) == 4 and well[3]:
                            s = "    %s%s AS %s," % (fo, s, fn)
                            name_changed += 1
                        else:
                            s = "    %s," % fn
                        lines.append(s)

                lines[-1] = lines[-1][:-1]  # kill the last comma
            else:
                raise Exception("fas should not be empty")

            if tfrom is None:
                tfrom = self.table

            if "\n" in tfrom:
                lines.append(
                    "FROM (\n%s) AS temp_tbl%d" %
                    (tfrom, self._count_as))
                self._count_as += 1
            else:
                lines.append("FROM %s" % tfrom)

            if len(where) > 0:
                wh = db._build_where_condition(where)
                lines.append(wh)

            return "\n".join(lines)

        def _find_in_fas(self, fas, a, b):
            """find a,b in fas (column 1 and 2)
            @param      fas     list [ (new_name, table, name)]
            @param      a       table name
            @param      b       name
            """
            for name, tbl, field in fas:
                if a == tbl and b == field:
                    return name
            raise ValueError(  # pragma: no cover
                "unable to find field %s.%s in (%s)" %
                (a, b, str(fas)))

        def _build_join(self, db, fas, select, n):
            """
            see :meth:`build_sql <pyensae.sql.database_join_group.build_sql>`

            @param      db              database
            @param      fas             list [(new_name, table, name)]
            @param      select          condition
            @param      n               node
            """

            other_select = n.SELECT
            parent_key = n.parent_key
            key = n.key

            other_select = other_select.split("\n")
            other_select = ["    " + s for s in other_select]
            other_select = "\n".join(other_select)
            select = select.split("\n")
            select = ["    " + s for s in select]
            select = "\n".join(select)

            res = self._build_select(db, fas, {}, select)
            res += "\nINNER JOIN (\n"
            res += other_select + ")"
            res += "\nON  "

            ppref = n.predecessor.PREFIX
            pref = n.PREFIX

            if isinstance(parent_key, str):
                if parent_key.startswith("<PREFIX>"):
                    a, b = parent_key[8:].split(".")
                    parent_key = (self._find_in_fas(fas, a, b),)
                else:
                    parent_key = (ppref + parent_key,)

                if key.startswith("<PREFIX>"):
                    a, b = key[8:].split(".")
                    key = (self._find_in_fas(fas, a, b),)
                else:
                    key = (pref + key,)

            else:
                pk = []
                k = []
                for m, n in zip(parent_key, key):
                    if m.startswith("<PREFIX>"):
                        a, b = m[8:].split(".")
                        pk.append(self._find_in_fas(fas, a, b))
                    else:
                        pk.append(ppref + m)

                    if n.startswith("<PREFIX>"):
                        a, b = n.split(".")
                        k.append(self._find_in_fas(fas, a, b))
                    else:
                        k.append(pref + n)

                parent_key = tuple(pk)
                key = tuple(k)

            oni = []
            for k, l in zip(parent_key, key):
                oni.append("%s == %s" % (k, l))
            oni = " AND \n    ".join(oni)
            res += oni
            return res

        def build_sql(self, db):
            """
            build the sql request

            @param      db      database

            The function adds two attributes:
                - SELECT:   sql request for a node
                - FIELDS:   list of [ (final_name, table, original_name)
            """
            self.check_prefix()

            for n in self.successor:
                n.build_sql(db)

            fas, where = self._in_select(db)
            fields = [f[:3] for f in fas]
            select = self._build_select(db, fas, where)

            if self.get_nb_successor() == 0:
                pass
            else:
                for n in self.successor:
                    fields.extend(n.FIELDS)
                    select = self._build_join(db, fields, select, n)

            self.SELECT = select
            self.FIELDS = fields

    def __init__(self):
        """
        constructor
        """
        self._count_as = 0

    ##################################################
    # the class itself: multiple joins using this tree
    ##################################################

    def inner_joins(self, root, execute=False, create_index=False, created_table=None,
                    duplicate_column=True, order=None, unique=False, distinct=False,
                    fields=None, nolog=True):
        """create several SQL inner join requests (included into each others)
        @param      root                JoinTreeNode (the root)
        @param      execute             if True, execute the query
        @param      create_index        if True, creates an index on the second table if it does not exist: it accelerates the inner join
        @param      created_table       if execute is True, you must specify a table name to be created
        @param      duplicate_column    do not include columns from the second table if their name is already in the first one
        @param      order               order clause, list of 2-tuple (column, way)  way is None or DESC
        @param      unique              unique or not
        @param      distinct            add the keyword DISTINCT
        @param      fields              restriction to fields given by fields or no restriction if None
        @param      nolog               if True, do not log the query
        @return                         SQL request, list of fields ("source", "new name")

        @warning Some options are not available yet:
                    - create_index      True
                    - duplicate_column  False
                    - order             != []
                    - unique            True

        @todo   Three tasks (however, this won't probably happen)
                    - Finish The function inner_joins (parameters create_index, duplicate_column, order, unique).
                    - Improve the handling of keyword DISTINCT
                    - Handle keyword fields
        """
        if order is None:
            order = []
        if create_index:
            raise RuntimeError(  # pragma: no cover
                "create_index = True: this option is not available")
        if not duplicate_column:
            raise RuntimeError(  # pragma: no cover
                "duplicate_column = False: this option is not available")
        if len(order) > 0:
            raise RuntimeError(  # pragma: no cover
                "order != []: this option is not available")
        if unique:
            raise RuntimeError(  # pragma: no cover
                "unique = True: this option is not available")
        if fields is not None:
            raise RuntimeError(  # pragma: no cover
                "fields != None: this option is not possible yet %s." %
                (str(fields)))

        root.build_sql(self)
        select = root.SELECT
        fields = root.FIELDS

        if distinct:
            if not select.startswith("SELECT"):
                raise ValueError("algorithm problem")  # pragma: no cover
            select = "SELECT DISTINCT" + select[len("SELECT"):]

        if execute:
            if created_table is None:
                raise RuntimeError(  # pragma: no cover
                    "unable to execute the SQL query: not specified name for the table to create")
            if created_table in self.get_table_list():
                raise ValueError("table %r already exists" % created_table)

            select = "CREATE TABLE %s AS \n" % created_table + select
            self.execute(select, nolog=nolog)

        return select, fields

    ##################################################
    # the other methods
    ##################################################

    def _build_where_condition(self, where, add_keyword_where=True):
        """builds a where condition (including the WHERE keyword)
        @param      where               condition where to interpret
                                            @code
                                            { "field": ("==", value) }
                                            @endcode
        @param      add_keyword_where   add the keyword where ?
        @return                         sql syntax

        @todo This function should deal with a tree to express AND and OR logical links.
              (However, this probably won't happen.)
        """
        sql = ""
        if where is not None and len(where) > 0:
            if add_keyword_where:
                sql += " WHERE "
            if isinstance(where, str):
                sql += where
            elif isinstance(where, dict):
                a = []
                for k, v in where.items():
                    if v[1] not in ['==', '<=', '>=', '>', '<', '!=']:
                        v = (v[1], v[0])
                    if v[1] not in ['==', '<=', '>=', '>', '<', '!=']:
                        raise ValueError(  # pragma: no cover
                            "unable to understand where %s,%s " %
                            (k, str(v)))
                    if v[1] == '==' and self.isMSSQL():
                        v = (v[0], '=')
                    v = (v[0], " %s " % v[1])
                    if isinstance(v[0], str):
                        if "'" in v[0]:
                            s = k + v[1] + "'" + v[0].replace("'", "''") + "'"
                        else:
                            s = k + v[1] + "'" + v[0] + "'"
                    else:
                        s = k + v[1] + str(v[0])
                    a.append(s)
                sql += " AND ".join(a)
            else:
                raise ValueError(  # pragma: no cover
                    "unable to interpret this where condition %s" %
                    (str(where)))
        return sql

    def histogram(self, table, columns, col_sums=None, values=None, sql_add=None,
                  execute=False, created_table=None, new_column="histogram",
                  nolog=False):
        """
        create a SQL request to compute an histogram

        @param      table               table
        @param      columns             column or columns (in a tuple) to be histogrammized
        @param      col_sums            candidate columns for a sum
        @param      values              specific values, several cases:
                                            - if None: does a GROUP BY
                                            - if dictionary of tuple:  ``{'cat1':('val1', 'val2', ...) }``
                                                then groups together several values into one category
                                            - if list of float: does an histogram on a real variable
        @param      new_column          name of the new column
        @param      sql_add             string to be added at the end of the SQL request
        @param      execute             if True, execute the request
        @param      created_table       the histogram can be stored into a table whose name is given by this parameter
        @param      nolog               if True, do not log the query
        @return                         SQL request

        """
        if col_sums is None:
            col_sums = []
        if isinstance(columns, str):
            columns = (columns,)

        cols = self.get_table_columns_list(table)
        for column in columns:
            if column not in [x[0] for x in cols]:
                raise ValueError(
                    "%r is not a column of table %r\n- columns:\n%r" %
                    (column, table, "\n".join(
                        [
                            str(x) for x in cols])))

        if sql_add is None or len(sql_add) == 0:
            sql_add = ""
        else:
            sql_add = ",\n       " + sql_add

        sum_column = []
        for c in col_sums:
            s = "SUM(%s) AS sum_%s" % (c, c)
            sum_column.append(s)
        str_sum = ", ".join(sum_column)
        if len(str_sum) > 0:
            str_sum = ", " + str_sum

        if values is None:
            sql = "SELECT %s AS %s, COUNT(%s) AS %s_nb%s%s\nFROM %s\nGROUP BY %s" % \
                (", ".join(columns),
                 new_column,
                 "*",
                 new_column,
                 str_sum,
                 sql_add,
                 table,
                 ", ".join(columns))
            select = sql

        elif isinstance(values, dict):
            values_rev = {}
            for k, vv in values.items():
                for v in vv:
                    if v not in values_rev:
                        values_rev[v] = []
                    values_rev[v].append(k)
            for k, v in values_rev.items():
                if len(v) > 1:
                    raise ValueError(  # pragma: no cover
                        "a category is shared by several values %r and %r" %
                        (k, ", ".join(v)))
            for k in values_rev:
                values_rev[k] = values_rev[k][0]

            def filterfunctionhistogramdict1(v):
                return values_rev.get(v, "none")

            def filterfunctionhistogramdict2(a, b):
                return values_rev.get((a, b), "none")

            def filterfunctionhistogramdict3(a, b, c):
                return values_rev.get((a, b, c), "none")

            def filterfunctionhistogramdict4(a, b, c, d):
                return values_rev.get((a, b, c, d), "none")

            def filterfunctionhistogramdict5(a, b, c, d, e):
                return values_rev.get((a, b, c, d, e), "none")

            self.add_function(
                "filterfunctionhistogramdict1",
                1,
                filterfunctionhistogramdict1)
            self.add_function(
                "filterfunctionhistogramdict2",
                2,
                filterfunctionhistogramdict2)
            self.add_function(
                "filterfunctionhistogramdict3",
                3,
                filterfunctionhistogramdict3)
            self.add_function(
                "filterfunctionhistogramdict4",
                4,
                filterfunctionhistogramdict4)
            self.add_function(
                "filterfunctionhistogramdict5",
                5,
                filterfunctionhistogramdict5)

            st = ",".join(["a", "b", "c", "d", "e"][:len(cols)])
            sql = "\n    -- def filterfunctionhistogramdict%d (%s) : return %s.get (%s, 'none')\n\n" % (
                len(columns), st, str(values_rev), st)
            sql += "\n    SELECT " + \
                ",\n           ".join([x[0] for x in cols])
            sql += ",\n           filterfunctionhistogramdict%d (%s) AS histo_temp_col" % (
                len(columns), ", ".join(columns),)
            sql += "\n    FROM %s" % table
            sql = "(" + sql + ") AS temp_tbl%d" % self._count_as
            self._count_as += 1

            select = "SELECT histo_temp_col AS %s,COUNT(histo_temp_col) AS %s_nb\n  %s\n  %s\nFROM %s\nGROUP BY histo_temp_col" % \
                (new_column, new_column, str_sum, sql_add, sql)

        elif isinstance(values, list):
            values = sorted(copy.copy(values))
            values2 = values[1:] + [max(1e10, max(values) + 1), ]
            names = list(values)
            couple = list(zip(range(0, len(values)), values, values2, names))

            def filterfunctionhistogramlist(v):
                for i, x, x_, n in couple:
                    if v < x_:
                        return n
                raise RuntimeError(  # pragma: no cover
                    "unable to process, " +
                    str(v) +
                    " is a value higher than 1e10")
            self.add_function(
                "filterfunctionhistogramlist",
                1,
                filterfunctionhistogramlist)

            sql = ""
            sql += "\n    SELECT " + ",\n    ".join([x[0] for x in cols])
            sql += ",\n    filterfunctionhistogramlist (%s) AS histo_temp_col" % (
                ", ".join(columns),)
            sql += "\n    FROM %s" % table
            sql = "(" + sql + ")"

            select = "SELECT histo_temp_col AS %s,COUNT(histo_temp_col) AS %s_nb\n  %s\n  %s\nFROM %s\nGROUP BY histo_temp_col" % \
                (new_column, new_column, str_sum, sql_add, sql)

        else:
            raise TypeError(  # pragma: no cover
                "values has not a type (%s) not in [None, dict, list]" %
                (str(
                    type(values))))

        if execute:
            if created_table is None:
                raise RuntimeError(  # pragma: no cover
                    "unable to execute the SQL query: not specified name for the table to create")
            if created_table in self.get_table_list():
                raise RuntimeError(  # pragma: no cover
                    "table %r already exists" % created_table)

            select = "CREATE TABLE %s AS \n" % created_table + select
            self.execute(select, nolog=nolog)

        return select

    def inner_join(self, table1, table2, field1, field2=None, where=None, execute=False,
                   create_index=True, created_table=None, prefix="", duplicate_column=True,
                   prefix_all="", order=None, unique=True, params=None, nolog=True):
        """create a SQL inner join request
        @param      table1              first table
        @param      table2              second table
        @param      field1              inner join on field1 from table1
        @param      field2              inner join on field2 from table2 (if None --> field2 = field1
        @param      where               where clause (if None, do not add it), dictionary or string
        @param      execute             if True, execute the query
        @param      create_index        if True, creates an index on the second table if it does not exist: it accelerates the inner join
        @param      created_table       if execute is True, you must specify a table name to be created
        @param      prefix              prefix for fields from the second table
        @param      duplicate_column    do not include columns from the second table if their name is already in the first one
        @param      prefix_all          prefix for all fields
        @param      order               order clause, list of 2-tuple (column, way)  way is None or DESC
        @param      unique              unique or not
        @param      params              special parameters for inner_joins method
        @param      nolog               if True, do not log the query, otherwise, skip that part
        @return                         SQL request, list of fields ("source", "new name")
        """
        if order is None:
            order = []
        if params is None:
            params = {}
        if field2 is None:
            field2 = field1

        cols1 = self.get_table_columns_list(table1)
        cols1 = [f[0] for f in cols1]
        if len(cols1) == 0:
            raise ValueError(  # pragma: no cover
                "table %r has no field" % table1)

        joinsm = table2 == '________________'
        if joinsm:
            cols2 = [f[1] for f in params["fields"]]
            if len(cols2) == 0:
                raise RuntimeError(  # pragma: no cover
                    "imported table has no field")
            table2 = params["sql"].split("\n")
            table2 = ["     " + s for s in table2]
            table2 = "\n".join(table2)
            table2 = "(%s)" % (table2.strip("\n\r "),)
        else:
            cols2 = self.get_table_columns_list(table2)
            cols2 = [f[0] for f in cols2]
            if len(cols2) == 0:
                raise Exception("table %s has no field" % table2)

        if isinstance(field1, tuple):
            for k in field1:
                if k not in cols1:
                    raise ValueError(  # pragma: no cover
                        "unable to find field %r in table %r" %
                        (k, table1))
            for k in field2:
                if k not in cols2:
                    raise ValueError(  # pragma: no cover
                        "unable to find field %r in table %r" %
                        (k, table2))
        else:
            if field1 not in cols1:
                raise ValueError(  # pragma: no cover
                    "unable to find field %r in table %r" %
                    (field1, table1))
            if field2 not in cols2:
                raise ValueError(  # pragma: no cover
                    "unable to find field %r in table %r" %
                    (field2, table2))
            field1 = (field1,)
            field2 = (field2,)

        if create_index and joinsm:
            li = self.get_index_list()
            ind = False
            for name, tbl_name, sql in li:
                if tbl_name != table2:
                    continue
                fields = re.compile("[(](\\w*)[)]").search(sql).groups()
                if len(fields) == 0:
                    continue
                field = fields[0]
                if field not in field2:
                    continue
                ind = True

            if not ind:
                self.LOG(  # pragma: no cover
                    "creating an index on table %r, field %r" %
                    (table2, ", ".join(field2)))
                self.create_index("index_" + table2.replace(".", "_") + "_" + "_".join(field2),
                                  table2, field2, unique=unique)

        keyfields = {}
        for k in field1:
            if k in cols2:
                if k not in field2:
                    keyfield = ("%s.%s" % (table1, k), table1 + "_" + k)
                else:
                    keyfield = ("%s.%s" % (table1, k), k)
            else:
                keyfield = (k, k)
            keyfields[k] = keyfield

        if "." in table1:
            ptable1 = table1.split(".")[1]
        else:
            ptable1 = table1
        if "." in table2:
            ptable2 = table2.split(".")[1]
        else:
            ptable2 = table2

        fields = []
        for c in cols1:
            if ":" in c:
                continue
            if c in keyfields:
                fields.append(keyfields[c])
            elif c in cols2:
                if duplicate_column:
                    fields.append(
                        ("%s.%s" %
                         (ptable1, c), "%s_%s" %
                            (ptable1, c)))
                else:
                    fields.append(("%s.%s" % (ptable1, c), c))
            else:
                fields.append((c, c))

        for c in cols2:
            if ":" in c:
                continue
            if c in field2:
                continue
            if c in cols1:
                if duplicate_column:
                    fields.append(
                        ("%s.%s" %
                         (ptable2, c), prefix + "%s_%s" %
                            (ptable2, c)))
            else:
                fields.append((c, prefix + c))

        mx = max([len(f[0]) for f in fields]) + 1
        rem = params.get("as_remove", None)

        if rem is None:
            fields = [(f[0] + " " * (mx - len(f[0])), prefix_all + f[1])
                      for f in fields]
        else:
            cfields = fields
            fields = []
            for f in cfields:
                a = f[0] + " " * (mx - len(f[0]))
                b = (prefix_all + f[1]).replace(rem, "")
                if b in cols1:
                    b = (prefix_all + f[1]).replace(rem, "_")
                fields.append((a, b))

        all = [" AS ".join(f) for f in fields]
        select = ",\n       ".join(all)
        select = "SELECT " + select + "\nFROM %s\n" % table1

        if unique:
            select += "INNER JOIN %s\n" % table2
        else:
            select += "JOIN %s\n" % table2

        nb = 0
        for k1, k2 in zip(field1, field2):
            if nb > 0:
                select += "   AND "
            else:
                select += "ON     "

            if k1 in cols2:
                select += "%s.%s " % (table1, k1)
            else:
                select += "%s " % k1

            if k2 in cols1 and not joinsm:
                select += "== %s.%s\n" % (table2, k2)
            else:
                select += "== %s\n" % k2

            nb += 1

        if where is not None and len(where) > 0:
            select += "WHERE " + where + "\n"

        if order is not None and len(order) > 0:
            te = []
            for o in order:
                if isinstance(o, tuple):
                    if o[0] in cols1 and o[0] in cols2:
                        te.append(ptable1 + "." + o[0] + " " + o[1])
                    else:
                        te.append(o[0] + " " + o[1])
                else:
                    if o in cols1 and o in cols2:
                        te.append(ptable1 + "." + o)
                    else:
                        te.append(o)
            select += "ORDER BY " + ", ".join(te) + "\n"

        #select += ";"

        if execute:
            if created_table is None:
                raise RuntimeError(  # pragma: no cover
                    "unable to execute the SQL query: not specified name for the table to create")
            if created_table in self.get_table_list():
                raise ValueError(  # pragma: no cover
                    "table %r already exists" % created_table)

            select = "CREATE TABLE %s AS \n" % created_table + select
            self.execute(select, nolog=nolog)

        fields = [(a_.strip(), b_) for a_, b_ in fields]
        return select, fields



"""
@file
@brief Shortcuts to file_helper
"""

from .content_helper import replace_comma_by_point
from .decompress_helper import decompress_zip, decompress_targz, decompress_gz
from .content_helper import file_head, file_tail, enumerate_grep, file_encoding



"""
@file
@brief Pandas and SQL
"""
import os
import pandas
from .database_main import Database


def import_flatfile_into_database_pandas(filedb, filetext, table=None,
                                         engine='SQLite', host='localhost', add_key=None,
                                         fLOG=print, **options):
    """
    Function which imports a file into a database using pandas.
    It the table exists, it removes it first. There is no addition.

    @param  filedb      something.db3
    @param  filetext    something.txt or .tsv
    @param  table       table name (in the database), if None, the database name will be the filename without extension
    @param  engine      engine to use when using a SQL server (SQLite or ODBCMSSQL)
    @param  host        host (server)
    @param  fLOG        logging function (will display information through the command line)
    @param  add_key     name of a key to add (or None if nothing to add)
    @param  encoding    encoding
    @param  options     options passed to `pandas.read_csv <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html>`_
    @return             table name
    """
    db = Database(filedb, engine=engine, host=host, LOG=fLOG)
    db.connect()

    if table is None:
        table = os.path.splitext(
            os.path.split(filetext)[-1])[0].replace(".", "").replace(",", "")

    if db.has_table(table):
        fLOG("remove ", table)
        db.remove_table(table)

    params = options.copy()
    params['filepath_or_buffer'] = filetext
    params['iterator'] = True
    params["chunksize"] = options.get('chunksize', 1000000)

    nb = 0
    for part in pandas.read_csv(**params):
        nb += part.shape[0]
        fLOG("number of added lines", nb)
        part.to_sql(con=db._connection, name=table,
                    if_exists="append", index=add_key)

    db.close()
    return table



"""
@file
@brief Downloads stock prices (from Yahoo website) and other prices.
"""
import os
import urllib.request
import urllib.error
import datetime
from io import StringIO
import pandas
import numpy
import requests
from pyquickhelper.filehelper import is_file_string


class StockPricesException(Exception):
    """
    Raised by StockPrices classes.
    """
    pass


class StockPricesHTTPException(StockPricesException):
    """
    Raised by StockPrices classes.
    """
    pass


class StockPrices:

    """
    Defines a class containing stock prices, provides basic functions,
    the class uses :epkg:`pandas` to load the data.

    .. exref::
        :title: Retrieve stock prices from the Yahoo source

        ::

            from pyensae.finance import StockPrices
            prices = StockPrices(tick="NASDAQ:MSFT")
            print(prices.dataframe.head())

    The class loads a stock price from either a url or a folder
    where the data was cached. If a filename
    ``<folder>/<tick>.<day1>.<day2>.txt`` already exists,
    it takes it from here. Otherwise, it downloads it.

    A couple of providers have been implemented but it is not
    easy to keep them up to date as policies from website
    change on a regular basis.
    If *url* is ``'yahoo'``, the data will be download using
    `CAC 40 <http://finance.yahoo.com/q/cp?s=^FCHI+Components>`_.
    The CAC40 composition is described by
    `Wikipedia CAC 40 <http://fr.wikipedia.org/wiki/CAC_40>`_.
    However `Yahoo Finance <https://fr.finance.yahoo.com/>`_
    introduced the use of cookies in May 2017
    and it is not so easy to automate.
    The default provider could be
    *Google Finance* which has now been integrated into the
    search engine.
    Tick names depends on the data prodiver. More details:
    `European Markets Information <https://www.stockmarketeye.com/users-guide/ticker-symbols-and-data-providers/euro-stocks.html>`_.
    You can also go to `quandl <https://www.quandl.com/data/EURONEXT/BNP-Bnp-Paribas-Act-A-BNP>`_
    and get the tick for the module `quandl <https://www.quandl.com/tools/python>`_.
    As of May 14th, the following error appears when using
    ``url='yahoo'`` which comes from an error in
    :epkg:`pandas_reader`::

        ImmediateDeprecationError(DEP_ERROR_MSG.format('Yahoo Daily'))
        pandas_datareader.exceptions.ImmediateDeprecationError:
        Yahoo Daily has been immediately deprecated due to large breaks in the API without the
        introduction of a stable replacement. Pull Requests to re-enable these data
        connectors are welcome.

        See https://github.com/pydata/pandas-datareader/issues

    ``url='yahoo_new'`` should solve the issue.
    It relies on :epkg:`yahoo_historial`.
    Data can be downloaded for a specific period of time.
    If not specified, it takes the largest available.

    .. exref::
        :title: Compute the average returns and correlation matrix

        ::

            import pyensae, pandas
            from pyensae.finance import StockPrices
            from pyensae.datasource import download_data

            # download the CAC 40 composition from my website (for Yahoo)
            download_data('cac40_2013_11_11.txt', website='xd')

            # download all the prices (if not already done) and store them into files
            actions = pandas.read_csv("cac40_2013_11_11.txt", sep="\\t")

            # we remove stocks with not enough historical data
            stocks = { k:StockPrices(tick = k) for k,v in actions.values }
            dates = StockPrices.available_dates(stocks.values())
            stocks = {k:v for k,v in stocks.items() if len(v.missing(dates)) <= 10}
            print("nb left", len(stocks))

            # we remove dates with missing prices
            dates = StockPrices.available_dates(stocks.values())
            ok = dates[dates["missing"] == 0]
            print("all dates before", len(dates), " after:" , len(ok))
            for k in stocks:
                stocks[k] = stocks[k].keep_dates(ok)

            # we compute correlation matrix and returns
            ret, cor = StockPrices.covariance(stocks.values(), cov = False, ret = True)

    You should also look at
    `pyensae et notebook <http://www.xavierdupre.fr/blog/notebooks/example%20pyensae.html>`_.
    If you use `Google Finance <https://www.google.com/finance>`_
    as a provider, the tick name is usually
    prefixed by the market places (NASDAQ for example). The export
    does not work for all markets places.
    Another provider was added, ``yahoo_new`` which delegates the task
    of getting data from `Yahoo Finance <https://finance.yahoo.com/>`_ to module
    `yahoo-historical <https://github.com/AndrewRPorter/yahoo-historical>`_.
    """

    def __init__(self, tick, url="google", folder="cache",
                 begin=None, end=None, sep=",",
                 intern=False, use_dtime=False):
        """
        @param      tick        tick name, ex ``NASDAQ:MSFT``
        @param      url         if yahoo, downloads the data from there if it was not done before
                                url is possible, ``'google'``, ``'yahoo_new'``,
                                ``'quandl'`` are predefined values
        @param      folder      cache folder (created if it does not exists
        @param      begin       first day (datetime), see below
        @param      end         last day (datetime), see below
        @param      sep         column separator
        @param      intern      do not use unless you know what to do
                                (see :meth:`__getitem__ <pyensae.finance.astock.StockPrices.__getitem__>`)
        @param      use_dtime   if True, use DateTime instead of string
        """
        if isinstance(url, pandas.DataFrame):
            self.datadf = url
            self.tickname = tick
            if "Date" not in url.columns:
                raise StockPricesHTTPException(
                    "the dataframe does not contain any column 'Date': {0}".format(
                        ",".join(
                            _ for _ in url.columns)))
        elif isinstance(tick, str) and is_file_string(tick) and os.path.exists(tick):
            self.tickname = os.path.split(tick)[-1]
            with open(tick, "r") as f:
                for line in f.readlines():
                    if line.startswith('<!DOCTYPE html PUBLIC'):
                        raise StockPricesHTTPException(
                            "pandas cannot parse the file, check your have access to internet: " + str(tick))
                    break
            try:
                self.datadf = pandas.read_csv(tick, sep=sep)
            except Exception as e:
                with open(tick, "r") as t:
                    content = t.read()
                if "Firewall Authentication" in content:
                    raise StockPricesException(
                        "pandas cannot parse the file, check your have access to internet: " + str(tick)) from e
                raise
        else:
            if not os.path.exists(folder):
                try:
                    os.mkdir(folder)
                except PermissionError as e:
                    raise StockPricesException(("PermissionError, unable to create directory '{0}', " +
                                                "check you execute the program in a folder you have " +
                                                "permission to modify ({1})").format(folder, os.getcwd())) from e
            self.tickname = tick

            if begin is None:
                begin = datetime.datetime(2000, 1, 3)
            if end is None:
                now = datetime.datetime.now()
                end = now - datetime.timedelta(1)

            sbeg = begin.strftime("%Y-%m-%d")
            send = end.strftime("%Y-%m-%d")
            name = os.path.join(folder, tick.replace(":", "_").replace("/", "_").replace("\\\\", "_") +
                                ".{0}.{1}.txt".format(sbeg, send))

            date_format = None
            if not os.path.exists(name):
                if url == "google":
                    use_url = True
                    url_string = "https://finance.google.com/finance/historical?q={0}".format(
                        self.tickname)
                    url_string += "&startdate={0}&enddate={1}&output=csv".format(
                        begin.strftime('%b %d, %Y'), end.strftime('%b %d, %Y'))
                    url = url_string.replace(" ", "+").replace(",", "%2C")
                    date_format = "%b-%d-%Y"
                elif url == "quandl":
                    import quandl  # pylint: disable=C0415
                    df = quandl.get(
                        "EURONEXT/BNP", start_date=begin.strftime('%Y-%m-%d'), end_date=end.strftime('%Y-%m-%d'))
                    df.reset_index(drop=False).to_csv(
                        name, sep=sep, index=False)
                    use_url = False
                elif url == 'yahoo_new':
                    from yahoo_historical import Fetcher

                    dt = datetime.datetime(begin.year, begin.month, begin.day)
                    ts = dt.timestamp()
                    try:
                        data = Fetcher(tick, ts)
                    except TypeError as e:
                        raise TypeError(
                            f"Unable to fetch data with year={begin.year!r}, "
                            f"month={begin.month!r}, day={begin.day!r}.") from e
                    df = data.get_historical()
                    df.to_csv(name, sep=sep, index=False)
                    use_url = False
                elif url in ("yahoo", "google", "fred", "famafrench"):
                    import pandas_datareader.data as web  # pylint: disable=C0415
                    df = web.DataReader(self.tickname, url,
                                        begin, end).reset_index(drop=False)
                    df.to_csv(name, sep=sep, index=False)
                    use_url = False
                else:
                    raise StockPricesHTTPException(
                        "Unable to download data '{0}' from the following website '{1}'".format(tick, url))

                if use_url:
                    self.url_ = url
                    try:
                        u = urllib.request.urlopen(url)
                        text = u.read()
                        u.close()
                    except urllib.error.HTTPError as e:
                        raise StockPricesHTTPException(
                            "HTTPError, unable to load tick '{0}'\nURL: {1}".format(tick, url)) from e

                    if len(text) < 10:
                        raise StockPricesHTTPException(
                            "nothing to download for '{0}' less than 10 downloaded bytes".format(tick))

                    try:
                        f = open(name, "wb")
                        f.write(text)
                        f.close()
                    except PermissionError as e:
                        raise StockPricesException(("PermissionError, unable to create directory '{0}', " +
                                                    "check you execute the program in a folder you have " +
                                                    "permission to modify ({1})").format(folder, os.getcwd())) from e
                else:
                    self.url_ = name

            try:
                self.datadf = pandas.read_csv(name, sep=sep)
            except Exception as e:
                with open(tick, "r") as t:
                    content = t.read()
                if "Firewall Authentication" in content:
                    raise StockPricesException(
                        "pandas cannot parse the file, check your have access to internet '{0}'".format(tick)) from e
                raise

            if date_format is not None:
                self.datadf["Date"] = pandas.to_datetime(self.datadf["Date"])
                self.datadf["Date"] = self.datadf["Date"].apply(
                    lambda x: x.strftime('%Y-%m-%d'))
                self.datadf.to_csv(name, sep=sep, index=False)

        if use_dtime:
            self.datadf["Date"] = pandas.to_datetime(self.datadf["Date"])

        if not intern:
            try:
                self.datadf = self.datadf.sort_values("Date")
            except ValueError as e:
                if "'Date' is both an index level and a column label" in str(e):
                    vals = self.datadf['Date']
                    ind = self.datadf.index
                    if numpy.array_equal(vals, ind):
                        self.datadf = self.datadf.sort_index()
                    else:
                        raise StockPricesException(
                            "Columns Date and index are different.") from e
                else:
                    raise
            except AttributeError:
                self.datadf = self.datadf.sort("Date")
            except KeyError as e:
                raise StockPricesException("schema: {}".format(
                    ",".join(self.datadf.columns))) from e
            self.datadf.reset_index(drop=True, inplace=True)
            self.datadf.set_index("Date", drop=False, inplace=True)

    def __getitem__(self, key):
        """
        Overloads the ``getitem`` operator to get a @see cl StockPrice object.

        @param      key     key
        @return             StockPrice
        """
        return StockPrices(
            self.tick, self.datadf.__getitem__(key), intern=True)

    def __len__(self):
        """
        @return     number of observations
        """
        return len(self.datadf)

    @property
    def shape(self):
        """
        @return     number of observations
        """
        return self.datadf.shape

    @property
    def tick(self):
        """
        Returns the tick name.
        """
        return self.tickname

    @property
    def dataframe(self):
        """
        Returns the dataframe.
        """
        return self.datadf

    def df(self):
        """
        Returns the dataframe.
        """
        return self.datadf

    def FirstDate(self):
        """
        Returns the first date.
        """
        return self.datadf["Date"].min()

    def LastDate(self):
        """
        Returns the first date.
        """
        return self.datadf["Date"].max()

    def missing(self, trading_dates):
        """
        Returnq the list of missing dates from an overset of trading dates.

        @param      trading_dates       trading_dates (DataFrame having the column ``Date`` or in the index)
        @return                         missing dates (or None if issues)
        """
        da = self.dataframe["Date"]
        da2 = {v: 1 for v in da}

        if isinstance(trading_dates, dict):
            se = trading_dates
        else:
            se = trading_dates[
                "Date"] if "Date" in trading_dates.columns else trading_dates.index

        tbl = [{"Date": v} for v in se if v not in da2]
        if len(tbl) > 0:
            df = pandas.DataFrame(tbl)
            try:
                return df.sort_values("Date")
            except AttributeError:
                return df.sort("Date")
        else:
            return None

    @staticmethod
    def available_dates(listStockPrices, missing=True, field="Close"):
        """
        Returns the list of values (Open or High or Low or Close or Volume) from each stock
        for all the available_dates for a list of stock prices.

        A missing date is a date for which there is at least one stock price and one missing stock price.

        if ``missing`` is true a column is added which gives the number of missing stock prices for this dates

        @param      listStockPrices     list of StockPrices
        @param      missing             True or False
        @param      field               which field to use to fill the matrix
        @return                         matrix with the available dates for each stock
        """
        if field == "ohlc":
            field = ["Open", "High", "Low", "Close"]
        dates = []
        if isinstance(field, str):
            for st in listStockPrices:
                lifi = list(st.dataframe.columns)
                index = lifi.index(field)
                for row in st.dataframe.values:
                    date = row[0]
                    dates.append(
                        {"Date": date, "tick": st.tick, field: row[index]})
        elif isinstance(field, (tuple, list)):
            for st in listStockPrices:
                lifi = list(st.dataframe.columns)
                indexes = [lifi.index(f) for f in field]
                for row in st.dataframe.values:
                    date = row[0]
                    r = {"Date": date, "tick": st.tick, }
                    for i, f in zip(indexes, field):
                        r[f] = row[i]
                    dates.append(r)
        else:
            raise TypeError("field must be a string, a tuple or a list")

        df = pandas.DataFrame(dates)
        if isinstance(field, str):
            piv = df.pivot("Date", "tick", field)
        elif isinstance(field, (tuple, list)):
            pivs = [df.pivot("Date", "tick", f) for f in field]
            for fi, piv in zip(field, pivs):
                col = [c + "," + fi for c in piv.columns]
                piv.columns = col
            if len(pivs) == 1:
                piv = pivs[0]
            else:
                piv = pivs[0].merge(pivs[1], how="outer",
                                    left_index=True, right_index=True)
                for p in pivs[2:]:
                    piv = piv.merge(
                        p, how="outer", left_index=True, right_index=True)
        else:
            raise TypeError("field must be a string, a tuple or a list")

        if missing:
            def count_nan(row):
                "count nans"
                n = 0
                for k, v in row.items():
                    if k == "Date":
                        continue
                    if numpy.isnan(v):
                        n += 1
                return n
            piv["missing"] = piv.apply(lambda row: count_nan(row), axis=1)

        try:
            piv = piv.sort_index()
        except AttributeError:
            piv = piv.sort()
        return piv

    def head(self):
        """
        usual
        """
        return self.dataframe.head()

    def tail(self):
        """
        usual
        """
        return self.dataframe.tail()

    def keep_dates(self, trading_dates):
        """
        removes undesired dates

        @param      trading_dates   dates
        @return                     new series
        """
        da = self.dataframe["Date"]
        da2 = {v: 1 for v in da}

        if isinstance(trading_dates, dict):
            se = trading_dates
        else:
            se = trading_dates[
                "Date"] if "Date" in trading_dates.columns else trading_dates.index

        tbl = {v: 1 for v in se if v in da2}
        if len(tbl) > 0:
            ave = self.dataframe.apply(lambda row: row["Date"] in tbl, axis=1)
            return StockPrices(self.tickname, self.dataframe.loc[ave, :])
        else:
            raise StockPricesException("no trading dates left")

    def returns(self):
        """
        Builds the series of returns.

        @param      col     column to use to compute the returns
        @return             StockPrices
        """
        df = self.dataframe
        fd = self.FirstDate()
        ld = self.LastDate()

        plus = df["Date"] > fd    # dates from FirstDate+1 to LastDate
        moins = df["Date"] < ld    # dates from FirstDate to LastDate-1

        res = df.loc[plus, ["Date", "Volume"]]

        for k in df.columns:
            if k in ["Date", "Volume"]:
                continue
            m = numpy.array(df.loc[moins, k])
            p = numpy.array(df.loc[plus, k])
            res[k] = (p - m) / m

        return StockPrices(self.tickname, res)

    @staticmethod
    def covariance(
            listStockPrices, missing=True, field="Close", cov=True, ret=False):
        """
        Computes the covariances matrix (of returns).

        @param      listStockPrices     list of StockPrices
        @param      field               which field to use to fill the matrix
        @param      cov                 if True, returns the covariance, otherwise, the correlations
        @param      ret                 if True, also add the returns
        @return                         square dataframe or 2 dataframe (returns, correlation)
        """
        listStockPrices = [v.returns() for v in listStockPrices]
        mat = StockPrices.available_dates(listStockPrices, False, field)

        npmat = numpy.array(mat)
        cov = numpy.cov(
            npmat.transpose()) if cov else numpy.corrcoef(
            npmat.transpose())
        names = [v.tick for v in listStockPrices]
        ret_mat = pandas.DataFrame(cov, columns=names, index=names)

        if ret:
            rows = [{"tick": v.tick, "return": v.dataframe[field].mean()}
                    for v in listStockPrices]
            ret = pandas.DataFrame(rows)
            ret.set_index("tick", drop=True, inplace=True)
            return ret, ret_mat
        else:
            return ret_mat

    def plot(self, begin=None, end=None,
             field="Close", date_format=None,
             existing=None, axis=1, ax=None,
             label_prefix=None, color=None, **args):
        """
        See :meth:`draw <pyensae.finance.astock.StockPrices.draw>`.
        """
        return StockPrices.draw(self, begin=begin, end=end,
                                field=field, date_format=date_format,
                                existing=existing, axis=axis, ax=ax,
                                label_prefix=label_prefix, color=color,
                                **args)

    @staticmethod
    def draw(listStockPrices, begin=None, end=None,
             field="Close", date_format=None,
             existing=None, axis=1, ax=None,
             label_prefix=None, color=None, **args):
        """
        Draws a graph showing one or several time series.
        The example was taken
        `date_demo.py <https://matplotlib.org/examples/api/date_demo.html>`_.

        @param      listStockPrices     list of @see cl StockPrices (or one @see cl StockPrices if it is the only one)
        @param      begin               first date (datetime) or None to take the first one
        @param      end                 last included date (datetime) or None to take the last one
        @param      field               Open, High, Low, Close, Adj Close, Volume
        @param      date_format         ``%Y`` or ``%Y-%m`` or ``%Y-%m-%d`` or None if you prefer the function to choose
        @param      args                other arguments to send to ``plt.subplots``
        @param      axis                1 or 2, it only works if existing is not None.
                                        If axis is 2, the function draws the curves on the second axis.
        @param      label_prefix        to prefix curve label
        @param      color               curve color
        @param      args                other parameters to give method ``plt.subplots``
        @param      ax                  use existing `axes <http://matplotlib.org/api/axes_api.html>`_
        @return                         `axes <http://matplotlib.org/api/axes_api.html>`_

        The parameter ``figsize`` of the method
        `subplots <https://matplotlib.org/api/pyplot_api.html?highlight=subplots#matplotlib.pyplot.subplots>`_
        can change the graph size (see the example below).

        .. exref::
            :title: graph of a financial series

            ::

                from pyensae.finance import StockPrices
                stocks = [ StockPrices("NASDAQ:MSFT", folder = cache),
                           StockPrices("NASDAQ:GOOGL", folder = cache),
                           StockPrices("NASDAQ:AAPL", folder = cache)]
                fig, ax, plt = StockPrices.draw(stocks)
                fig.savefig("image.png")
                fig, ax, plt = StockPrices.draw(stocks, begin="2010-01-01", figsize=(16,8))
                plt.show()

            You can also chain the graphs and add a series on a second graph:

            ::

                from pyensae.finance import StockPrices
                stock = StockPrices("NASDAQ:MSFT", folder = cache)
                stock2 = StockPrices "NASDAQ:GOOGL", folder = cache)
                fig, ax, plt = stock.plot(figsize=(16,8))
                fig, ax, plt = stock2.plot(existing=(fig,ax), axis=2)
                plt.show()

        .. versionchanged:: 1.1
            Parameter *existing* was removed and parameter *ax* was added.
            If the date overlaps, the method
            `autofmt_xdate <https://matplotlib.org/api/figure_api.html#matplotlib.figure.Figure.autofmt_xdate>`_
            should be called.
        """
        if isinstance(listStockPrices, StockPrices):
            listStockPrices = [listStockPrices]

        data = StockPrices.available_dates(
            listStockPrices, missing=False, field=field)
        if begin is None:
            if end is not None:
                data = data[data.index <= end]
        else:
            if end is not None:
                data = data[(data.index >= begin) & (data.index <= end)]
            else:
                data = data[data.index >= begin]

        dates = [datetime.datetime.strptime(_, '%Y-%m-%d') for _ in data.index]
        begin = dates[0]
        end = dates[-1]

        def price(x):
            "local formatting"
            return '%1.2f' % x

        import matplotlib.pyplot as plt  # pylint: disable=C0415
        import matplotlib.dates as mdates  # pylint: disable=C0415

        if ax is not None:
            ex_h, ex_l = ax.get_legend_handles_labels()
            ex_l = tuple(ex_l)
            ex_h = tuple(ex_h)
            if axis == 2:
                ax = ax.twinx()
            fig = None
        else:
            if 'label' in args:
                args_ = {k: v for k, v in args.items() if k not in ('label', )}
            else:
                args_ = args
            fig, ax = plt.subplots(**args_)
            ex_h, ex_l = tuple(), tuple()

        curve = []
        if field == "ohlc":
            from mplfinance.original_flavor import candlestick_ohlc  # pylint: disable=E0401
            ohlc = list(list(data.iloc[i, :4])
                        for i in range(0, data.shape[0]))
            ohlc = [[mdates.date2num(t)] + v for t, v in zip(dates, ohlc)]
            candlestick_ohlc(ax, ohlc, colorup="g")
        else:
            if label_prefix is None:
                label_prefix = ""
            add_args = {}
            if color:
                add_args['c'] = color
            for stock in data.columns:
                if axis == 2:
                    curve.append(
                        ax.plot(dates, data[stock], "r", linestyle='solid',
                                label=label_prefix + str(stock), **add_args))
                else:
                    curve.append(
                        ax.plot(dates, data[stock], linestyle='solid', c=color,
                                label=label_prefix + str(stock), **add_args))

        if existing is None:
            ax.format_xdata = mdates.DateFormatter('%Y-%m-%d')
            if len(dates) < 30:
                days = mdates.DayLocator()
                ax.xaxis.set_major_locator(days)
                ax.xaxis.set_minor_locator(days)
                if date_format is not None:
                    fmt = mdates.DateFormatter(date_format)
                    ax.xaxis.set_major_formatter(fmt)
                else:
                    ax.xaxis.set_major_formatter(
                        mdates.DateFormatter("%Y-%m-%d"))
            elif len(dates) < 500:
                months = mdates.MonthLocator()
                days = mdates.DayLocator()
                ax.xaxis.set_major_locator(months)
                ax.xaxis.set_minor_locator(days)
                ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y-%m"))
                if date_format is not None:
                    fmt = mdates.DateFormatter(date_format)
                    ax.xaxis.set_major_formatter(fmt)
                else:
                    ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y-%m"))
            else:
                years = mdates.YearLocator()
                months = mdates.MonthLocator()
                ax.xaxis.set_major_locator(years)
                ax.xaxis.set_minor_locator(months)
                if date_format is not None:
                    fmt = mdates.DateFormatter(date_format)
                    ax.xaxis.set_major_formatter(fmt)
                else:
                    ax.xaxis.set_major_formatter(mdates.DateFormatter("%Y"))

        ax.set_xlim(begin, end)
        ax.format_ydata = price
        if fig is not None:
            fig.autofmt_xdate()

        if axis == 2:
            if isinstance(curve, list):
                curve = [_[0] for _ in curve]
            ax.legend(ex_h + tuple(curve), ex_l + tuple(data.columns))
        else:
            ax.grid(True)
            ax.legend(ex_l + tuple(data.columns))

        return ax

    def to_csv(self, filename, sep="\t", index=False, **params):
        """
        Saves the file in text format,
        see `to_csv <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html>`_

        @param      filename        filename
        @param      sep             separator
        @param      index           to keep or drop the index
        @param      params          other parameters
        """
        self.dataframe.to_csv(filename, sep=sep, index=index, **params)

    def to_excel(self, excel_writer, **params):
        """
        Saves the file in Excel format,
        see `to_excel <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_excel.html>`_
        """
        self.dataframe.to_excel(excel_writer, **params)



# -*- coding: utf-8 -*-
"""
@file
@brief Abstract class to connect to a SQL server using various way.
It will be used to implement magic functions
"""
import pandas
from pyquickhelper.loghelper import noLOG
from .database_main import Database
from .sql_interface import InterfaceSQL, InterfaceSQLException


class InterfaceSQLDatabase(InterfaceSQL):

    """
    Abstract class to connect to a SQL server using various way.
    It will be used to implement magic functions
    """

    def __init__(self, filename):
        """
        @param      filename        str or :class:`Database <pyensae.sql.database_main.Database>`

        If *filename* is a :class:`Database <pyensae.sql.database_main.Database>`,
        the object is kept as is.

        .. versionchanged:: 1.1
            Parameter *filename* can be a database.
        """
        if isinstance(filename, Database):
            self.obj = filename
        else:
            self.obj = Database(filename, LOG=noLOG)

    def connect(self):
        """
        connection to the database
        """
        self.obj.connect()
        self.populate_completion()

    def close(self):
        """
        close the connection to the database
        """
        self.obj.close()

    def get_table_list(self):
        """
        returns the list of tables in the database

        @return     list of strings
        """
        return self.obj.get_table_list()

    def get_table_columns(self, table_name, as_dict=True):
        """
        returns the list of columns in a table

        @param      table_name      table name
        @param      as_dict         True, as dictionary, as a list otherwise
        @return                     dictionary { "column": (position, type) }
        """
        return self.obj.get_table_columns(table_name, dictionary=as_dict)

    def execute_clean_query(self, sql_query):
        """
        return the resuls of a SQL query

        @param      sql_query       query to execute
        @return                     pandas DataFrame
        """
        con = self.obj._connection
        try:
            return pandas.read_sql(sql_query, con)
        except pandas.io.sql.DatabaseError:
            try:
                self.obj.execute_view(sql_query)
            except Exception as ee:
                raise InterfaceSQLException(str(ee)) from ee

    def import_flat_file(self, filename, table_name):
        """
        import a flat file as a table, we assume the columns
        separator is ``\\t`` and the file name contains a header

        @param      filename        filename
        @param      table           table name
        @return                     the number of added rows
        """
        r = self.obj.import_table_from_flat_file(
            filename,
            table_name,
            columns=None,
            header=True)
        self.populate_completion()
        return r

    def drop_table(self, table_name):
        """
        drops a table

        @param      table           table name
        """
        self.obj.remove_table(table_name)
        self.populate_completion()

    def add_function(self, code_function):
        """
        add a function to the database which can be called in a SELECT statement

        @param      code_function  pointer to the function
        """
        name = code_function.__name__
        nbp = code_function.__code__.co_argcount
        self.obj.add_function(name, nbp, code_function)

    def import_dataframe(self, tablename, df):
        """
        import a dataframe into the database

        @param      tablename       name of the table
        @param      df              dataframe
        @return                     the number of added rows
        """
        df.to_sql(tablename, self.obj._connection)
        self.populate_completion()
        return len(df)



"""
@file
@brief Various functions to process text
"""

import os
import re
import chardet


def replace_comma_by_point(file):
    """
    Replaces all commas by point in a file (do that inplace).

    :param file: file to process
    """
    with open(file, "r") as f:
        text = f.read()
    text = text.replace(",", ".")
    with open(file, "w") as f:
        f.write(text)


def file_head(filename: str, nbline=10, encoding="utf8", errors="strict"):
    """
    Extracts the first *nbline* of a file (assuming it is text file).

    :param filename: filename
    :param nbline: number of lines
    :param encoding: encoding
    :param errors: see `open <https://docs.python.org/3/library/functions.html#open>`_
    :return: list of lines
    """
    if isinstance(filename, str):
        if not os.path.exists(filename):
            raise FileNotFoundError(filename)
        if not os.path.isfile(filename):
            raise FileNotFoundError(  # pragma: no cover
                "'{0}' is not a file".format(filename))
        with open(filename, "r", encoding=encoding, errors=errors) as f:
            return file_head(f, nbline=nbline, encoding=encoding)
    else:
        rows = []
        for line in filename:
            rows.append(line)
            if len(rows) >= nbline:
                break
        return rows


def file_tail(filename: str, nbline=10, encoding="utf8", threshold=2 ** 14, errors="strict"):
    """
    Extracts the first nbline of a file (assuming it is text file).

    :param filename: filename
    :param nbline: number of lines
    :param encoding: encoding
    :param threshold: if the file size is above, it will not read the beginning
    :param errors: see `open <https://docs.python.org/3/library/functions.html#open>`_
    :return: list of lines

    The line marked as *A* has an issue because the cursor
    could fall on a character (= byte) in the middle of a character
    if the file is encoded in utf-8 character.
    The next line fails. That's why we try again
    by moving the cursor by one character (see line B).

    The first returned line may be incomplete.
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(filename)  # pragma: no cover
    if not os.path.isfile(filename):
        raise FileNotFoundError(  # pragma: no cover
            "'{0}' is not a file".format(filename))

    size = os.stat(filename).st_size
    if size < threshold:
        with open(filename, "r", encoding=encoding, errors=errors) as f:
            rows = f.readlines()
        return rows[-nbline:] if len(rows) > nbline else rows
    else:
        with open(filename, "r", encoding=encoding, errors=errors) as f:
            f.seek(size - threshold)  # line A
            try:
                content = f.read()
            except UnicodeDecodeError:
                f.seek(size - threshold - 1)  # line B
                content = f.read()

        rows = content.split("\n")
        res = rows[-nbline:] if len(rows) > nbline else rows
        return [_ + "\n" for _ in res]


def enumerate_grep(filename, regex, encoding="utf8", errors=None):
    """
    Extracts lines matching a regular expression.

    @param      filename        filename
    @param      regex           regular expression
    @param      encoding        encoding
    @param      errors          see `open <https://docs.python.org/3/library/functions.html#open>`_
    @return                     iterator in lines

    .. versionadded:: 1.1
    """
    if isinstance(filename, str):
        if not os.path.exists(filename):
            raise FileNotFoundError(filename)  # pragma: no cover
        if not os.path.isfile(filename):
            raise FileNotFoundError(  # pragma: no cover
                "'{0}' is not a file".format(filename))
        with open(filename, "r", encoding=encoding, errors=errors) as f:
            for _ in enumerate_grep(f, regex, encoding):
                yield _
    else:
        reg = re.compile(regex)
        for line in filename:
            if reg.search(line):
                yield line


def file_encoding(filename_or_bytes, limit=2**20):
    """
    Returns the encoding of a file.
    The function relies on `chardet <http://chardet.readthedocs.io/en/latest/usage.html>`_.

    :param filename_or_bytes: filename or bytes
    :param limit: if *filename_or_bytes* is a file, the function only
        loads the first *limit* bytes (or all if *limit* is -1)
    :return: dictionary

    Example of results:

    ::

        {'encoding': 'EUC-JP', 'confidence': 0.99}
    """
    if isinstance(filename_or_bytes, str):
        if not os.path.exists(filename_or_bytes):
            raise FileNotFoundError(filename_or_bytes)
        if not os.path.isfile(filename_or_bytes):
            raise FileNotFoundError(
                "'{0}' is not a file".format(filename_or_bytes))
        size = os.stat(filename_or_bytes).st_size
        with open(filename_or_bytes, "rb") as f:
            content = f.read() if limit == -1 or size < limit else f.read(limit)
        return file_encoding(content)
    elif isinstance(filename_or_bytes, bytes):
        return chardet.detect(filename_or_bytes)
    else:
        raise TypeError("Unexpecting type for filename_or_bytes, got: {0}.".format(
            type(filename_or_bytes)))



"""
@file

@brief @see cl Database
"""

import os
import collections
try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable
from .file_text_binary import TextFile
from .file_text_binary_columns import TextFileColumns
from .database_exception import DBException


class DatabaseImportExport:

    """
    This class is not meant to be working alone.
    It contains import, export function for a database, in various formats.
    """

    ##########################################################################
    # exporting functions
    ##########################################################################

    def export_table_into_flat_file(self, table, filename, header=False, columns=None,
                                    post_process=None, encoding="utf8"):
        """
        Exports a table into a flat file.

        @param      table           table name
        @param      filename        filename
        @param      header          add a header on the first line
        @param      columns         export only columns in this list (if None, export all)
        @param      post_process    post_process a line:
                                        - input:  list, dictionary (for your own use, same one all the time)
                                        - output: list
        @param      encoding        encoding

        .. exref::
            :title: Export the results of a SQL query into a flat file
            :tag: SQL

            ::

                from pyensae.sql.database_main import Database
                dbfile = "filename.db3"
                filetxt = "fileview.txt"
                sql = "..."
                db = Database(dbfile)
                db.connect()
                db.export_view_into_flat_file (sql, fileview, header = True)
                db.close()
        """
        if columns is None:
            sql = "SELECT * FROM " + table + ";"
        else:
            sql = "SELECT %s FROM %s ;" % (",".join(columns), table)

        self.export_view_into_flat_file(
            sql, filename, header, post_process, encoding=encoding)

    def _clean_string(self, s):
        """
        Cleans string.

        @param      s       string
        @return             remove \\r\\t\\n
        """
        rep = {"\t": "\\t",
               "\n": "\\n",
               "\r": "\\r", }
        for k, v in rep.items():
            s = s.replace(k, v)
        return s

    def export_view_into_flat_file(self, view_sql, filename, header=False, post_process=None,
                                   encoding="utf8"):
        """
        Exports a table into a flat file.

        @param      view_sql        SQL request
        @param      filename        filename
        @param      header          if != None, add a header on the first line (header is a list of string)
        @param      post_process    if != None, use this function to post-process a text line extracted from the file
        @param      encoding        if != None, use this as a parameter to convert any value into str
        """
        sepline = "\n"

        self._check_connection()

        if header:
            if isinstance(header, (list, tuple)):
                header_line = "\t".join(header) + sepline
            elif isinstance(header, bool):
                col = self.get_sql_columns(view_sql)
                header_line = "\t".join(col) + sepline
            else:
                header_line = header + sepline
        else:
            header_line = ""

        sql = view_sql
        cur = self.execute(sql)
        nbline = 0

        f = open(filename, "w", encoding=encoding)
        f.write(header_line)
        memo = {}

        for line_ in cur:

            if post_process is not None:
                line = post_process(line_, memo)
            else:
                line = line_

            pr = "\t".join([self._clean_string(str(x)) for x in line])

            f.write(pr + sepline)
            nbline += 1
            if nbline % 100000 == 0:
                self.LOG("     exporting from view, line ", nbline)

        f.close()
        cur.close()

    ##########################################################################
    # importing functions
    ##########################################################################

    def append_values(self, values, tablename, schema, cursor=None,
                      skip_exception=False, encoding="utf-8"):
        """
        Uses @see me _append_table to fill a table will the
        values contained in values (as list).

        @param      values          list of list (each cell is a value)
        @param      tablename       name of the table to fill
        @param      schema          schema of the database, it must be present in case on the columns
                                    includes the tag "PRIMARYKEY", in that case, the value for this field
                                    will be automatically set up.
        @param      cursor          if None, create a new one
        @param      skip_exception  skip exception while inserting an element
        @param      encoding        encoding

        """
        self._append_table(
            values,
            tablename,
            schema,
            cursor=cursor,
            skip_exception=skip_exception,
            encoding=encoding)

    def _append_table(self, file, table, columns, format="tsv", header=False,
                      stop=-1, lower_case=False, cursor=None, fill_missing=0,
                      unique=None, filter_case=None, strict_separator=False,
                      skip_exception=False, changes=None, encoding="utf-8",
                      **params):
        """
        Appends element to a database.

        @param      file                file name or a matrix (this matrix can be an iterator)
        @param      table               table name
        @param      columns             columns definition (see below)
        @param      format              tsv, the only one accepted for the time being, it can be a function (line, **params)
        @param      header              the file has a header of not, if True, skip the first line
        @param      stop                if -1, insert every line, otherwise stop when the number of inserted lines is stop
        @param      lower_case          put every str string in lower_case before inserting it
        @param      cursor              if None, create a new one
        @param      fill_missing        fill the missing values by a default value, at least not more than fill_missing values
        @param      unique              if unique is a column number,
                                        the function will not take into account another containing a value already seen on this column
        @param      filter_case         process every case information (used to replace space for example)
        @param      strict_separator    strict number of columns, it assumes there is no separator in the content of every column
        @param      params              see format
        @param      skip_exception      skip exception while inserting an element
        @param      changes             to rewrite column names
        @param      encoding            encoding
        @return                         number of inserted elements

        The columns definition must follow the schema:
            - dictionary ``{ key:(column_name,python_type) }``
            - or ``{ key:(column_name,python_type,preprocessing_function) }``

        ``preprocessing_function`` is a function whose prototype is for example:

        @code
        def preprocessing_score (s) :
            return s.replace (",",".")
        @endcode

        And:
            - if ``PRIMARYKEY`` is added, the key is considered as the primary key
            - if ``AUTOINCREMENT`` is added, the key will automatically filled (like an id)

        """
        if changes is None:
            changes = {}
        if stop != -1:
            self.LOG("SQL append table stop is ", stop)
        self._check_connection()
        nbinsert = 0
        unique_key = {}
        if isinstance(file, list) or (
                isinstance(file, Iterable) and not isinstance(file, str)):
            primarykey = None
            for c, v in columns.items():
                if "PRIMARYKEY" in v:
                    primarykey = v[0]

            if table not in self.get_table_list():
                raise DBException("unable to find table " + table)

            all = 0
            num_line = 0
            for line in file:
                if stop != -1 and all >= stop:
                    break
                dic = self._process_text_line(line, columns, format=format, lower_case=lower_case,
                                              num_line=num_line, filter_case=filter_case,
                                              strict_separator=strict_separator)

                if unique is not None:
                    if dic[unique] in unique_key:
                        continue
                    unique_key[dic[unique]] = 0

                num_line += 1
                if dic is not None:
                    self._get_insert_request(dic, table, True, primarykey, cursor=cursor,
                                             skip_exception=skip_exception)
                    nbinsert += 1
                    #self._connection.execute (s)
                    all += 1
                    if all % 100000 == 0:
                        self.LOG(
                            "adding %d lines into table %s" %
                            (all, table))
        else:
            primarykey = None
            for c, v in columns.items():
                if "PRIMARYKEY" in v:
                    primarykey = v[0]

            if table not in self.get_table_list():
                table_list = self.get_table_list()
                message = "unable to find table " + table + \
                    " in [" + ",".join(table_list) + "]"
                raise DBException(message)

            column_has_space = len(
                [v[0] for k, v in columns.items() if ' ' in v[0]]) > 0
            self.LOG(
                "   column_has_space", column_has_space, [
                    v[0] for k, v in columns.items()])

            if strict_separator or column_has_space:
                file = TextFile(file, errors='ignore',
                                fLOG=self.LOG, encoding=encoding)
                skip = False
            else:
                self.LOG("   changes", changes)
                file = TextFileColumns(file, errors='ignore', fLOG=self.LOG,
                                       regex=columns, changes=changes, encoding=encoding)
                skip = True

            file.open()
            all = 0
            num_line = 0
            every = 100000
            tsv = format == "tsv"

            for line in file:
                if stop != -1 and all >= stop:
                    break
                num_line += 1
                if skip:
                    dic = line
                else:
                    if header and num_line == 1:
                        continue
                    if len(line.strip("\r\n")) == 0:
                        continue
                    if tsv:
                        dic = self._process_text_line(line, columns, format, lower_case=lower_case,
                                                      num_line=num_line - 1, fill_missing=fill_missing,
                                                      filter_case=filter_case, strict_separator=strict_separator)
                    else:
                        dic = format(line, **params)
                        if dic is None:
                            continue

                if unique is not None:
                    if dic[unique] in unique_key:
                        continue
                    unique_key[dic[unique]] = 0

                if dic is not None:
                    self._get_insert_request(
                        dic, table, True, primarykey, cursor=cursor)
                    nbinsert += 1
                    all += 1
                    if all % every == 0:
                        self.LOG(
                            "adding %d lines into table %s" %
                            (all, table))
            file.close()

        if cursor is not None:
            cursor.close()
        self.commit()
        return nbinsert

    def import_table_from_flat_file(self, file, table, columns, format="tsv", header=False,
                                    display=False, lower_case=False, table_exists=False,
                                    temporary=False, fill_missing=False, indexes=None,
                                    filter_case=None, change_to_text=None, strict_separator=False,
                                    add_key=None, encoding="utf-8", **params):
        """
        Adds a table to database from a file.

        @param      file                file name or matrix
        @param      table               table name
        @param      columns             columns definition (see below)
                                        if None: columns are guessed
        @param      format              tsv, the only one accepted for the time being,
                                        it can be a function whose parameter are a line and **params
        @param      header              the file has a header of not, if True, skip the first line
        @param      lower_case          put every string in lower case before inserting it
        @param      table_exists        if True, do not create the table
        @param      temporary           adding a temporary table
        @param      fill_missing        fill the missing values
        @param      indexes             add indexes before appending all the available observations
        @param      filter_case         process every case information (used to replace space for example)
        @param      encoding            encoding
        @param      params              see format
        @param      change_to_text      changes the format from any to TEXT
        @param      display             if True, print more information on stdout
        @param      strict_separator    strict number of columns, it assumes there is no separator in the content of every column
        @param      add_key             name of a key to add (or None if nothing to add)
        @return                         the number of added rows

        The columns definition must follow the schema:

        - dictionary ``{ key: (column_name,python_type) }``
        - or ``{ key: (column_name,python_type,preprocessing_function) }``

        ``preprocessing_function`` is a function whose prototype is for example:

        ::

            def preprocessing_score (s) :
                return s.replace (",",".")

        And:

        - if ``PRIMARYKEY`` is added, the key is considered as the primary key
        - if ``AUTOINCREMENT`` is added, the key will automatically filled (like an id)

        @warning The function does not react well when a column name
                 includes a space.
        """
        if indexes is None:
            indexes = []
        if change_to_text is None:
            change_to_text = []
        if display:
            if isinstance(file, list):
                self.LOG("processing file ", file[:min(len(file), 10)])
            else:
                self.LOG("processing file ", file)

        self._check_connection()
        if columns is None:
            # here, some spaces might have been replaced by "_", we need to get
            # them back
            columns, changes = self._guess_columns(
                file, format, columns, filter_case=filter_case, header=header, encoding=encoding)
        elif isinstance(columns, list):
            columns_, changes = self._guess_columns(
                file, format, columns, filter_case=filter_case, header=header, encoding=encoding)
            if len(columns_) != len(columns):
                raise DBException(
                    "different number of columns:\ncolumns={0}\nguessed={1}".format(
                        str(columns), str(columns_)))
            columns = columns_

        if add_key is not None:
            columns[len(columns)] = (
                add_key, int, "PRIMARYKEY", "AUTOINCREMENT")

        for i in columns:
            v = columns[i]
            if v[0] in change_to_text:
                if len(v) <= 2:
                    v = (v[0], (str, 1000000))
                else:
                    v = (v[0], (str, 1000000)) + v[2:]
            columns[i] = v

        if display:
            self.LOG("   columns ", columns)

        if not isinstance(file, list) and not os.path.exists(file):
            raise DBException("unable to find file " + file)

        if not table_exists:
            cursor = self.create_table(table, columns, temporary=temporary)
        elif table not in self.get_table_list():
            raise DBException("unable to find table " + table + " (1)")
        else:
            cursor = None

        if table not in self.get_table_list():
            raise DBException("unable to find table " + table + " (2)")
        nb = self._append_table(file, table, columns, format=format, header=header,
                                lower_case=lower_case, cursor=cursor, fill_missing=fill_missing,
                                filter_case=filter_case, strict_separator=strict_separator,
                                changes=changes, encoding=encoding, **params)

        self.LOG(nb, " lines imported")

        for ind in indexes:
            if isinstance(ind, str):
                indexname = table + "_" + ind
            else:
                indexname = table + "_" + "_".join(ind)
            if not self.has_index(indexname):
                self.create_index(indexname, table, ind)

        return nb
